-- **************************************************
-- Provide Moho with the name of this script object
-- **************************************************

ScriptName = "FA_BindPoints"

-- **************************************************
-- General information about this script
-- **************************************************

FA_BindPoints = {}

function FA_BindPoints:Name()
	return "Bind Points"
end

function FA_BindPoints:Version()
	return "5.0"
end

function FA_BindPoints:Description()
	return "Select/deselect points (press <space> to bind points, <ctrl> to override lasso mode)"
end

function FA_BindPoints:Creator()
	return "Lost Marble, Modified by Fazek"
end

function FA_BindPoints:UILabel()
	return("Bind Points")
end

-- **************************************************
-- The guts of this script
-- **************************************************

function FA_BindPoints:IsEnabled(moho)
	if (moho.frame > 0 or moho:CountPoints() < 1) then
		return false
	end
	if (moho:ParentSkeleton() == nil) then
		return false
	end
	return true
end

function FA_BindPoints:OnMouseDown(moho, mouseEvent)
	FA_SharedUtils.selectionMode= 0
	FA_SharedUtils:MouseDown(moho,mouseEvent)
end

function FA_BindPoints:OnMouseMoved(moho, mouseEvent)
	FA_SharedUtils:MouseMoved(moho, mouseEvent)
end

function FA_BindPoints:OnMouseUp(moho, mouseEvent)
	FA_SharedUtils:MouseUp(moho, mouseEvent)
end

function FA_BindPoints:OnKeyDown(moho, keyEvent)
	if (keyEvent.key == ' ') then
		self:HandleMessage(moho, keyEvent.view, self.BIND)
--[[
	elseif (keyEvent.keyCode == LM.GUI.KEY_UP) then
		FA_SelectBone:OnKeyDown(moho, keyEvent)
	elseif (keyEvent.keyCode == LM.GUI.KEY_DOWN) then
		FA_SelectBone:OnKeyDown(moho, keyEvent)
]]--
	end
end

-- **************************************************
-- Tool options - create and respond to tool's UI
-- **************************************************

FA_BindPoints.BIND = MOHO.MSG_BASE
FA_BindPoints.LASSO = MOHO.MSG_BASE + 1
FA_BindPoints.DUMMY = MOHO.MSG_BASE + 2
FA_BindPoints.SELECTBONE = MOHO.MSG_BASE + 3
FA_BindPoints.SELECTITEM = MOHO.MSG_BASE + 300

function FA_BindPoints:DoLayout(moho, layout)
	self.boneMenu = LM.GUI.Menu(FA_SharedUtils.txSelectBone)

	self.bonePopup = LM.GUI.PopupMenu(128, false)
	self.bonePopup:SetMenu(self.boneMenu)
	layout:AddChild(self.bonePopup)

	self.pointMenu = LM.GUI.Menu(FA_SharedUtils.txSelectGroup)

	self.pointPopup = LM.GUI.PopupMenu(128, false)
	self.pointPopup:SetMenu(self.pointMenu)
	layout:AddChild(self.pointPopup)

	self.bindBut = LM.GUI.Button(FA_SharedUtils.txBindPoints, self.BIND)
	layout:AddChild(self.bindBut)

	self.lassoCheck = LM.GUI.CheckBox(FA_SharedUtils.txLasso, self.LASSO)
	layout:AddChild(self.lassoCheck)
end

function FA_BindPoints:UpdateWidgets(moho)
	local skel = moho:ParentSkeleton()
	if (skel == nil) then
		return
	end
	local mesh = moho:Mesh()
	if (mesh == nil) then
		return
	end

	MOHO.BuildBoneMenu(self.boneMenu, skel, self.SELECTBONE, self.DUMMY)
	MOHO.BuildGroupMenu(self.pointMenu, mesh, self.SELECTITEM, self.DUMMY)
	self.lassoCheck:SetValue(FA_SharedUtils.lassoMode)
end

function FA_BindPoints:HandleMessage(moho, view, msg)
	local skel = moho:ParentSkeleton()
	if (skel == nil) then
		return
	end

	local mesh = moho:Mesh()
	if (mesh == nil) then
		return
	end

	if (msg == self.BIND) then
		moho.document:PrepUndo(moho.layer)
		moho.document:SetDirty()

		local boneID = -1
		for i = 0, skel:CountBones() - 1 do
			if (skel:Bone(i).fSelected) then
				boneID = i
				break
			end
		end
		for i = 0, mesh:CountPoints() - 1 do
			local pt = mesh:Point(i)
			if (pt.fSelected) then -- attach selected points
				pt.fParent = boneID
			elseif (pt.fParent == boneID) then -- detach unselected points that used to be attached
				pt.fParent = -1
			end
		end
	elseif (msg == self.LASSO) then
		FA_SharedUtils.lassoMode = self.lassoCheck:Value()
	elseif (msg >= self.SELECTBONE and msg < self.SELECTITEM) then
		FA_SharedUtils:SelectBone(moho, skel, msg - self.SELECTBONE, true)
	elseif (msg >= self.SELECTITEM) then
		FA_SharedUtils:SelectItem(moho, mesh, msg - self.SELECTITEM)
	end
end
