-- **************************************************
-- Provide Moho with the name of this script object
-- **************************************************

ScriptName = "FA_Curvature"

-- **************************************************
-- General information about this script
-- **************************************************

FA_Curvature = {}

function FA_Curvature:Name()
	return "Curvature"
end

function FA_Curvature:Version()
	return "5.0"
end

function FA_Curvature:Description()
	return "Drag side to side, or drag the handles to adjust the curvature, or set it numerically"
end

function FA_Curvature:Creator()
	return "Lost Marble, Modified by Fazek"
end

function FA_Curvature:UILabel()
	return("Curvature")
end

function FA_Curvature:LoadPrefs(prefs)
	FA_SharedUtils:Load_Prefs(prefs)
end

function FA_Curvature:SavePrefs(prefs)
	FA_SharedUtils:Save_Prefs(prefs)
end

-- **************************************************
-- Recurring values
-- **************************************************

FA_Curvature.selList = nil
FA_Curvature.savedVal = 0
FA_Curvature.focusID= -1

-- **************************************************
-- The guts of this script
-- **************************************************

function FA_Curvature:IsEnabled(moho)
	if (moho:CountPoints() > 0) then
		return true
	end
	return false
end

function FA_Curvature:OnMouseDown(moho, mouseEvent)
	local mesh = moho:Mesh()
	if (mesh == nil) then
		return
	end
	self.focusID= -1

	if (mouseEvent.shiftKey) then
		FA_SharedUtils.selectionMode= 0
		FA_SharedUtils:MouseDown(moho, mouseEvent)
	else
		if (moho:CountSelectedPoints() > 0) then
			local ptID= mouseEvent.view:PickPoint(mouseEvent.pt)

			if (ptID >= 0) then

				if (mesh:Point(ptID).fSelected) then
					moho.document:PrepUndo(moho.layer)
					moho.document:SetDirty()
					self:SetSelList(mesh, ptID)

					self.focusID= ptID
					FA_SharedUtils.mouseMode= 0
					mouseEvent.view:DrawMe()
					return
				end
			end
		end
		FA_SharedUtils.selectionMode= 1
		FA_SharedUtils:MouseDown(moho, mouseEvent)

		if (FA_SharedUtils.mouseMode >= 0 and moho:CountSelectedPoints() < 1) then
			FA_SharedUtils.selectionMode= 0
		elseif (FA_SharedUtils.mouseMode == -1) then
			self:SetNumeric(moho)
		end
	end
end

function FA_Curvature:OnMouseMoved(moho, mouseEvent)
	local mesh = moho:Mesh()
	if (mesh == nil) then
		return
	end

	if (FA_SharedUtils.mouseMode ~= 0) then
		FA_SharedUtils:MouseMoved(moho, mouseEvent)

		if (FA_SharedUtils.mouseMode == 0) then
			moho.document:PrepUndo(moho.layer)
			moho.document:SetDirty()
			self:SetSelList(mesh, -1)
			self.savedVal = 0
		else
--waiting for movement
			return
		end
	end

--mouse drag

	if (moho:CountSelectedPoints() < 1) then
		return
	end
	local newVal = (mouseEvent.pt.x - mouseEvent.startPt.x) / mouseEvent.view:Graphics():Width()
	local cVal= 0
	local pt, cve

	for i= 1, table.getn(self.selList) do
		cve= self.selList[i][1]
		local ptPos,c

		if (cve == nil) then
--all curves through the point
			pt= mesh:Point(self.selList[i][2])
			ptPos= -1
			local cveN= pt:CountCurves() - 1

			for j= 0, cveN do
				cve, ptPos = pt:Curve(j, ptPos)
				c = cve:GetCurvature(ptPos, moho.frame)
				c = c - self.savedVal + newVal
				c = LM.Clamp(c, MOHO.PEAKED, 2 / 3)
				cve:SetCurvature(ptPos, c, moho.frame)
			end

			if (cveN == 0) then
				cVal= c
			end
		else
--only in the selected directions
			ptPos= self.selList[i][2]
			c = cve:GetCurvature(ptPos, moho.frame)
			c = c - self.savedVal + newVal
			c = LM.Clamp(c, MOHO.PEAKED, 2 / 3)
			cve:SetCurvature(ptPos, c, moho.frame)
			cVal= c
		end
	end

	if (cVal ~= 0 and table.getn(self.selList) < 2) then
		self.cvText:SetValue(cVal)
	end
	self.savedVal = newVal
	mouseEvent.view:DrawMe()
end

function FA_Curvature:OnMouseUp(moho, mouseEvent)

	if (FA_SharedUtils.mouseMode > 0) then
		FA_SharedUtils:MouseUp(moho, mouseEvent)
		self:SetNumeric(moho)
		return
	end
	moho:NewKeyframe(CHANNEL_CURVE)
	self.selList = nil
	mouseEvent.view:DrawMe()
end

function FA_Curvature:OnKeyDown(moho, keyEvent)
	FA_SharedUtils:KeyDown(moho, keyEvent)
end

function FA_Curvature:DrawMe(moho, view)
	local gfx = view:Graphics()
	local matrix = LM.Matrix:new_local()

	moho.layer:GetFullTransform(moho.frame, matrix, moho.document)
	gfx:Push()
	gfx:ApplyMatrix(matrix)

	if (self.focusID >= 0) then
		local mesh = moho:Mesh()

		if (mesh ~= nil) then
			local pt= mesh:Point(self.focusID)

			if (pt.fSelected) then
				gfx:SetColor(255, 0, 0)
				gfx:DrawFatMarker(pt.fPos.x, pt.fPos.y, 6)
			end
		end
	end
	view:DrawPreviewShape()
	gfx:Pop()
end

function FA_Curvature:SetNumeric(moho)
-- set the numeric display to the curvature of the selected point

	if (moho:CountSelectedPoints() < 2) then

		local mesh = moho:Mesh()

		if (mesh == nil) then
			return
		end

		for i= 0, mesh:CountPoints() - 1 do
			local pt= mesh:Point(i)

			if (pt.fSelected and pt:CountCurves() == 1) then
				local cve
				local cvePt= -1
				cve, cvePt= pt:Curve(0,cvePt)
				self.cvText:SetValue(cve:GetCurvature(cvePt, moho.frame))
				return
			end
		end
	end
end

function FA_Curvature:SetSelList(mesh,ptID)

	self.selList= {}

	local pt,cve

	if (ptID >= 0) then

--a specially selected (focus) point

		pt= mesh:Point(ptID)

	--modify this single point, only the selected curves (or all curves if no neighbouring selected points)

		local cvePt= -1

		for i= 0, pt:CountCurves() - 1 do
			cve, cvePt = pt:Curve(i, cvePt)
			local cveN= cve:CountPoints() - 1
			local f= false

			if (cvePt == 0) then

				if (cve.fClosed) then
					f= cve:Point(cveN).fSelected
				end
			else
				f= cve:Point(cvePt - 1).fSelected
			end

			if (cvePt == cveN) then

				if (cve.fClosed) then
					f= (f or cve:Point(0).fSelected)
				end
			else
				f= (f or cve:Point(cvePt + 1).fSelected)
			end

			if (f) then
				table.insert(self.selList, { cve, cvePt })
			end
		end

		if (table.getn(self.selList) == 0) then
			table.insert(self.selList, { nil, ptID })
		end
	else
--all selected points

		for i= 0, mesh:CountPoints() - 1 do
			pt= mesh:Point(i)

			if (pt.fSelected) then
		--clear markers
				pt.fTempPos.x= 0
			end
		end

		for i= 0, mesh:CountCurves() - 1 do
			cve= mesh:Curve(i)
			local cveN= cve:CountPoints() - 1
			pt= cve:Point(0)
			local fPrev= false
			local fLast= false

			if (cve.fClosed) then
				fPrev= cve:Point(cveN).fSelected
				fLast= pt.fSelected
			end

			for j= 1, cveN do
				local next= cve:Point(j)

				if (pt.fSelected) then

					if (fPrev) then
						table.insert(self.selList, { cve, j - 1 })
						pt.fTempPos.x= 1
					else
						if (j < cveN) then
							if (next.fSelected) then
								table.insert(self.selList, { cve, j - 1 })
								pt.fTempPos.x= 1
							end
						end
					end
					fPrev= true
				else
					fPrev= false
				end
				pt= next
			end

		--last point on the curve

			if (pt.fSelected and (fPrev or fLast)) then
				table.insert(self.selList, { cve, cveN })
				pt.fTempPos.x= 1
			end
		end

	--add remaining points: all connected curves

		for i= 0, mesh:CountPoints() - 1 do
			pt= mesh:Point(i)

			if (pt.fSelected and (pt.fTempPos.x == 0)) then
				table.insert(self.selList, { nil, i })
			end
		end
	end
end

function FA_Curvature:SetCurvatures(moho, mesh, cVal)

--set the curvatures to a value, and refresh

	self:SetSelList(mesh, self.focusID)
	local sLen= table.getn(self.selList)

	if (sLen < 1) then return end

	local cve

	moho.document:PrepUndo(moho.layer)
	moho.document:SetDirty()

	for i= 1, sLen do
		cve= self.selList[i][1]
		local ptPos

		if (cve == nil) then
--all curves through the point
			local pt= mesh:Point(self.selList[i][2])
			ptPos= -1
			local cveN= pt:CountCurves() - 1

			for j= 0, cveN do
				cve, ptPos = pt:Curve(j, ptPos)
				cve:SetCurvature(ptPos, cVal, moho.frame)
			end
		else
--only in the selected directions
			ptPos= self.selList[i][2]
			cve:SetCurvature(ptPos, cVal, moho.frame)
		end
	end
	self.selList= nil
	moho:NewKeyframe(CHANNEL_CURVE)
end

-- **************************************************
-- Tool options - create and respond to tool's UI
-- **************************************************

FA_Curvature.DUMMY = MOHO.MSG_BASE
FA_Curvature.MODIFY = MOHO.MSG_BASE + 1
FA_Curvature.PEAK = MOHO.MSG_BASE + 2
FA_Curvature.SMOOTH = MOHO.MSG_BASE + 3
FA_Curvature.LASSO = MOHO.MSG_BASE + 4
FA_Curvature.SELECTITEM = MOHO.MSG_BASE + 5

function FA_Curvature:DoLayout(moho, layout)
	self.menu = LM.GUI.Menu("Select Group")

	self.popup = LM.GUI.PopupMenu(128, false)
	self.popup:SetMenu(self.menu)
	layout:AddChild(self.popup)

	layout:AddChild(LM.GUI.StaticText("Curvature:"))

	self.cvText= LM.GUI.TextControl(0, "000.0000", 0, LM.GUI.FIELD_FLOAT)
	self.cvText:SetWheelInc(0.05)
	self.cvText:SetValue(0)
	layout:AddChild(self.cvText)

	self.modifyBut= LM.GUI.Button(FA_SharedUtils.txModify, self.MODIFY)
	layout:AddChild(self.modifyBut)
	self.peakBut= LM.GUI.Button("Peak", self.PEAK)
	layout:AddChild(self.peakBut)
	self.smoothBut= LM.GUI.Button("Smooth", self.SMOOTH)
	layout:AddChild(self.smoothBut)

	self.lassoCheck = LM.GUI.CheckBox(FA_SharedUtils.txLasso, self.LASSO)
	layout:AddChild(self.lassoCheck)

	self.focusID= -1
	self.selList= nil
end

function FA_Curvature:UpdateWidgets(moho)
	self.lassoCheck:SetValue(FA_SharedUtils.lassoMode)
	local en= false

	if (moho:CountSelectedPoints() > 0) then
		en= true
	end
	self.modifyBut:Enable(en)
	self.peakBut:Enable(en)
	self.smoothBut:Enable(en)

	local mesh = moho:Mesh()
	if (mesh == nil) then
		return
	end
	MOHO.BuildGroupMenu(self.menu, mesh, self.SELECTITEM, self.DUMMY)
end

function FA_Curvature:HandleMessage(moho, view, msg)
	local mesh = moho:Mesh()
	if (mesh == nil) then
		return
	end

	if (msg >= self.SELECTITEM) then
		FA_SharedUtils:SelectItem(moho, mesh, msg - self.SELECTITEM)
	elseif (msg == self.MODIFY) then
		self:SetCurvatures(moho, mesh, self.cvText:FloatValue())

	elseif (msg == self.PEAK) then
		self:SetCurvatures(moho, mesh, MOHO.PEAKED)

	elseif (msg == self.SMOOTH) then
		self:SetCurvatures(moho, mesh, MOHO.SMOOTH)

	elseif (msg == self.LASSO) then
		FA_SharedUtils.lassoMode = self.lassoCheck:Value()
	end
end
