-- **************************************************
-- Provide Moho with the name of this script object
-- **************************************************

ScriptName = "FA_Freehand"

-- **************************************************
-- General information about this script
-- **************************************************

FA_Freehand = {}

function FA_Freehand:Name()
	return "Freehand"
end

function FA_Freehand:Version()
	return "5.0"
end

function FA_Freehand:Description()
	return "Freehand drawing (press <space> to set a peak point, hold <alt> to disable auto-welding)"
end

function FA_Freehand:Creator()
	return "Lost Marble, Modified by Fazek"
end

function FA_Freehand:UILabel()
	return("Freehand")
end

function FA_Freehand:LoadPrefs(prefs)
	FA_SharedUtils:Load_Prefs(prefs)
	self.pixelTolerance = prefs:GetInt("FA_Freehand.pixelTolerance", 10)
	self.angleTolerance = prefs:GetInt("FA_Freehand.angleTolerance", 10)
	self.widthMode = prefs:GetInt("FA_Freehand.widthMode", 1)
	self.minWidth = prefs:GetFloat("FA_Freehand.minWidth", 1)
	self.maxWidth = prefs:GetFloat("FA_Freehand.maxWidth", 8)
	self.taperStart = prefs:GetBool("FA_Freehand.taperStart", false)
	self.taperEnd = prefs:GetBool("FA_Freehand.taperEnd", false)
end

function FA_Freehand:SavePrefs(prefs)
	FA_SharedUtils:Save_Prefs(prefs)
	prefs:SetInt("FA_Freehand.pixelTolerance", self.pixelTolerance)
	prefs:SetInt("FA_Freehand.angleTolerance", self.angleTolerance)
	prefs:SetInt("FA_Freehand.widthMode", self.widthMode)
	prefs:SetFloat("FA_Freehand.minWidth", self.minWidth)
	prefs:SetFloat("FA_Freehand.maxWidth", self.maxWidth)
	prefs:SetBool("FA_Freehand.taperStart", self.taperStart)
	prefs:SetBool("FA_Freehand.taperEnd", self.taperEnd)
end

-- **************************************************
-- Recurring values
-- **************************************************

FA_Freehand.startPoint= 0
FA_Freehand.breakPoint= 0
FA_Freehand.lastX = 0
FA_Freehand.lastY = 0
FA_Freehand.lastDrawX = 0
FA_Freehand.lastDrawY = 0
FA_Freehand.firstVecX = 0.0
FA_Freehand.firstVecY = 0.0
FA_Freehand.previewWidth = 1
FA_Freehand.weldToPt= -1
FA_Freehand.weldToCurve= -1
FA_Freehand.weldToSeg= -1
FA_Freehand.disableTaperStart = false
FA_Freehand.disableTaperEnd = false
FA_Freehand.mouseon= false
FA_Freehand.firstPt= false

FA_Freehand.pixelTolerance = 10
FA_Freehand.angleTolerance = 10
FA_Freehand.widthMode = 1 -- 0: none, 1: pen pressure, 2: random, 3: constant
FA_Freehand.minWidth = 1
FA_Freehand.maxWidth = 8
FA_Freehand.taperStart = false
FA_Freehand.taperEnd = false

-- **************************************************
-- The guts of this script
-- **************************************************

function FA_Freehand:IsEnabled(moho)
	local mesh = moho:Mesh()
	if (mesh == nil) then
		return false
	end
	if (moho.layer:CurrentAction() ~= "") then
-- creating new objects in the middle of an action can lead to unexpected results
		return false
	end
	return true
end

function FA_Freehand:OnMouseDown(moho, mouseEvent)
	local mesh = moho:Mesh()
	if (mesh == nil) then return end

	self.startPoint = mesh:CountPoints()
	self.firstPt= true
	self.firstVecX= mouseEvent.vec.x
	self.firstVecY= mouseEvent.vec.y

	self.weldToPt= -1
	self.weldToCurve= -1
	self.weldToSeg= -1

	if ((not mouseEvent.altKey) and FA_SharedUtils.autoWeld) then

--find the closest point for auto-welding the freehand curve to

		local testVec1 = LM.Vector2:new_local()
		testVec1:Set(mouseEvent.vec)
		self.weldToPt= mesh:ClosestPoint(testVec1)

		if (self.weldToPt >= 0) then
			local testVec2 = LM.Vector2:new_local()
			testVec2:Set(mesh:Point(self.weldToPt).fPos)

	-- convert the two chosen points to pixel coordinates
	-- this is to make sure that they are actually close enough
	-- in screen space
			local p1 = LM.Point:new_local()
			local p2 = LM.Point:new_local()
			local m = LM.Matrix:new_local()
			moho.layer:GetFullTransform(moho.frame, m, moho.document)
			m:Transform(testVec1)
			m:Transform(testVec2)
			mouseEvent.view:Graphics():WorldToScreen(testVec1, p1)
			mouseEvent.view:Graphics():WorldToScreen(testVec2, p2)
			p1.x = p1.x - p2.x
			p1.y = p1.y - p2.y
			if ((p1.x * p1.x) + (p1.y * p1.y) >= 100) then self.weldToPt= -1 end
		end

		if (self.weldToPt < 0) then
--find a curve segment to auto-weld the freehand curve to
			self.weldToCurve,self.weldToSeg= mouseEvent.view:PickEdge(mouseEvent.pt,self.weldToCurve,self.weldToSeg)
		end
	end
	self.previewWidth = moho:PixelToDoc(1)
	self.lastX = mouseEvent.startPt.x
	self.lastY = mouseEvent.startPt.y
	self.lastDrawX = mouseEvent.startPt.x
	self.lastDrawY = mouseEvent.startPt.y
	self.mouseon= true
	self.breakPoint= self.startPoint
	mouseEvent.view:DrawMe()
end

function FA_Freehand:OnMouseMoved(moho, mouseEvent)
	local mesh = moho:Mesh()
	if (mesh == nil) then return end

	local ptWidth

	if (FA_SharedUtils.autoOutline and self.widthMode > 0) then
		ptWidth= 0.0

		if (self.widthMode == 1) then

			if (mouseEvent.penPressure > 0) then
				ptWidth= mouseEvent.penPressure
			end
		elseif (self.widthMode == 2) then
			ptWidth= math.random()
		end
		ptWidth= LM.Lerp(ptWidth,self.minWidth,self.maxWidth)
		ptWidth= moho:PixelToDoc(ptWidth)
	else
		ptWidth= -1
	end

-- move the most recent point to the current position, and set its width

	local	ptN= mesh:CountPoints()

	if (not self.firstPt) then
		local pt = mesh:Point(ptN - 1)
		pt.fPos:Set(mouseEvent.vec)
		pt.fAnimPos:SetValue(moho.frame, mouseEvent.vec)

		if (ptWidth >= 0) then
			pt.fWidth:SetValue(moho.frame, ptWidth)
		end
	end

-- test if the mouse has moved far enough for a new point

	local dX= mouseEvent.pt.x - self.lastX
	local dY= mouseEvent.pt.y - self.lastY

	if ((dX * dX) + (dY * dY) > self.pixelTolerance * self.pixelTolerance) then

		if (self.firstPt) then
--create first point
			moho.document:PrepUndo(moho.layer)
			moho.document:SetDirty()
			mesh:SelectNone()

			local firstVec= LM.Vector2:new_local()
			firstVec:Set(self.firstVecX,self.firstVecY)

			self.disableTaperStart = false
			self.disableTaperEnd = false

			if (FA_SharedUtils.autoClose) then
--closed curves
				mesh:AddLonePoint(firstVec,moho.frame)
				mesh:AppendPoint(mouseEvent.vec, moho.frame)

				if (FA_SharedUtils.sharpCorners) then
					mesh:Point(ptN):SetCurvature(MOHO.PEAKED,0)
					mesh:Point(ptN + 1):SetCurvature(MOHO.PEAKED,0)
				end

				self.weldToPt= -1
			else
				if (self.weldToPt >= 0) then
--attach to an existing point
					self.disableTaperStart = true
					mesh:AddLonePoint(firstVec,moho.frame)
					mesh:AppendPoint(mouseEvent.vec, moho.frame)

					if (FA_SharedUtils.sharpCorners) then
						mesh:Point(ptN):SetCurvature(MOHO.PEAKED,0)
						mesh:Point(ptN + 1):SetCurvature(MOHO.PEAKED,0)
					end

				elseif (self.weldToCurve >= 0) then
--add a point in the middle of some curve
					self.disableTaperStart = true
					mesh:AddPoint(firstVec, self.weldToCurve, self.weldToSeg, moho.frame)
					mesh:AddPoint(mouseEvent.vec, -1, moho.frame)

					if (FA_SharedUtils.sharpCorners) then
						mesh:Point(ptN):SetCurvature(MOHO.PEAKED,0)
						mesh:Point(ptN + 1):SetCurvature(MOHO.PEAKED,0)
					end
					mesh:WeldPoints(ptN + 1, ptN, moho.frame)
				else
--no auto-welding, just start a new curve
					mesh:AddLonePoint(firstVec, moho.frame)
					mesh:AppendPoint(mouseEvent.vec, moho.frame)

					if (FA_SharedUtils.sharpCorners) then
						mesh:Point(ptN):SetCurvature(MOHO.PEAKED,0)
						mesh:Point(ptN + 1):SetCurvature(MOHO.PEAKED,0)
					end
				end
			end
			self.firstPt= false
			pt= mesh:Point(ptN)
			pt.fWidth:SetValue(moho.frame, ptWidth)
		else
--add a new point
			mesh:AppendPoint(mouseEvent.vec, moho.frame)

			if (FA_SharedUtils.sharpCorners) then
				mesh:Point(ptN):SetCurvature(MOHO.PEAKED,0)
			end
		end
		pt= mesh:Point(mesh:CountPoints() - 1)
		pt.fWidth:SetValue(moho.frame, ptWidth)
		self.lastX = mouseEvent.pt.x
		self.lastY = mouseEvent.pt.y
		self.previewWidth = ptWidth
	end

-- preview
	local gfx = mouseEvent.view:Graphics()
	dX= (gfx:Height() / moho.document:Height()) * 0.85
	dX= moho:DocToPixel(self.previewWidth) * gfx:CurrentScale(false) * dX 

	gfx:SetSmoothing(true)
	gfx:Push()
	local m = gfx:CurrentTransform()
	m:Invert()
	gfx:ApplyMatrix(m)
	gfx:SetColor(255, 0, 0)
	gfx:SetPenWidth(dX)

	gfx:MoveTo(self.lastDrawX, self.lastDrawY)
	gfx:LineTo(mouseEvent.pt.x, mouseEvent.pt.y)
	gfx:Pop()
	gfx:SetSmoothing(false)

	mouseEvent.view:Refresh()

	self.lastDrawX = mouseEvent.pt.x
	self.lastDrawY = mouseEvent.pt.y
end

function FA_Freehand:OnMouseUp(moho, mouseEvent)
	self.mouseon= false

	if (self.firstPt) then return end

	local mesh = moho:Mesh()
	if (mesh == nil) then
		mouseEvent.view:DrawMe()
		return
	end

	if (self.weldToPt >= 0) then
-- the initial point should be welded to an existing point
-- we do this after drawing the curve
		mesh:WeldPoints(self.startPoint, self.weldToPt, moho.frame)
	end

	local n = mesh:CountPoints()

	if (mesh:CountPoints() > self.startPoint + 2) then

-- the last two points may lie on top of each other - if so, delete one

		local v = mesh:Point(n - 1).fPos - mesh:Point(n - 2).fPos

		if (v:Mag() < 1e-6) then
			mesh:Point(n - 2).fSelected = true
			mesh:DeletePoint(n - 1)
		end
	end
	mesh:SelectNone()

	if (mesh:CountPoints() <= self.startPoint) then
		mouseEvent.view:DrawMe()
	end

	if (not(FA_SharedUtils.autoClose) and not(mouseEvent.altKey) and FA_SharedUtils.autoWeld) then
-- auto-weld the final point of the freehand curve

		if (self:WeldEndpoint(moho, mouseEvent.view, self.startPoint, mesh:CountPoints() - 1)) then

-- welded to an existing point
			self.disableTaperEnd = true
		else
-- try welding the endpoint to a curve segment other than this freehand curve

			local ptID = mesh:CountPoints() - 1
			local m = LM.Matrix:new_local()
			local v = LM.Vector2:new_local()
			local screenPt = LM.Point:new_local()
			local curveID = -1
			local segID = -1

			moho.layer:GetFullTransform(moho.frame, m, moho.document)
			v:Set(mesh:Point(ptID).fPos)
			m:Transform(v)
			mouseEvent.view:Graphics():WorldToScreen(v, screenPt)
			curveID, segID = mesh:Point(ptID):GetEndpointEdge(curveID, segID)
			curveID, segID = mouseEvent.view:PickEdge(screenPt, curveID, -1)
			if (curveID >= 0) then

	-- add a point in the middle of some curve

				if (not
				mesh:Curve(curveID):IsPointOnSegment(ptID,segID)) then
	-- don't weld the point back on itself
					mesh:AddPoint(mesh:Point(ptID).fPos,curveID,segID,moho.frame)
					mesh:WeldPoints(ptID, mesh:CountPoints() - 1,moho.frame)
					self.disableTaperEnd= true
				end
			end
		end
	end

	if (self.taperStart or self.taperEnd) then

-- taper the beginning and/or the end of the new line

		local f = moho:NewShapeLineWidth()
		local dist = 0
		local prevV = LM.Vector2:new_local()
		prevV:Set(mesh:Point(self.startPoint).fPos)

		for i = self.startPoint + 1, mesh:CountPoints() - 1 do
			local pt = mesh:Point(i)

			if (pt.fWidth.value < 0) then
				pt.fWidth:SetValue(moho.frame, f)
			end
			dist = dist + (pt.fPos - prevV):SquaredMag()
			prevV:Set(pt.fPos)
		end
		dist = dist / 4

		if (self.taperStart and not self.disableTaperStart) then
	-- beginning of the curve

			local n = 1
			local testDist = 0
			prevV:Set(mesh:Point(self.startPoint).fPos)
			for i = self.startPoint + 1, mesh:CountPoints() - 1 do
				local pt = mesh:Point(i)
				testDist = testDist + (pt.fPos - prevV):SquaredMag()
				prevV:Set(pt.fPos)
				if (testDist >= dist) then
					break
				end
				n = n + 1
			end
			for i = self.startPoint, self.startPoint + n - 1 do
				f = (i - self.startPoint) / n
				local pt = mesh:Point(i)
				pt.fWidth:SetValue(moho.frame, pt.fWidth.value * f)
			end
		end

		if (self.taperEnd and not self.disableTaperEnd) then
	-- end of the curve

			local numP = mesh:CountPoints()
			local n = 1
			local testDist = 0
			prevV:Set(mesh:Point(numP - 1).fPos)
			for i = numP - 2, self.startPoint, -1 do
				local pt = mesh:Point(i)
				testDist = testDist + (pt.fPos - prevV):SquaredMag()
				prevV:Set(pt.fPos)
				if (testDist >= dist) then
					break
				end
				n = n + 1
			end
			for i = numP - 1, numP - n, -1 do
				f = (numP - 1 - i) / n
				local pt = mesh:Point(i)
				pt.fWidth:SetValue(moho.frame, pt.fWidth.value * f)
			end
		end
	end

	if (self.angleTolerance > 0 and (mesh:CountPoints() > self.startPoint + 2)) then
-- adaptive freehand, remove "straight" sections of the curve

		local adaptiveLimit = math.cos(math.rad(self.angleTolerance))
		local runningVec = mesh:Point(self.startPoint + 1).fPos - mesh:Point(self.startPoint).fPos
		runningVec:NormMe()
		local i = self.startPoint + 2

		while (i < mesh:CountPoints()) do
			local testVec = mesh:Point(i).fPos - mesh:Point(i - 1).fPos
			testVec:NormMe()
			if (runningVec:Dot(testVec) > adaptiveLimit) then
				mesh:DeletePoint(i - 1)
				i = i - 1
			else
				runningVec = testVec
			end
			i = i + 1
		end
	end

	if (FA_SharedUtils.autoClose) then
		local farOutVec = LM.Vector2:new_local()
		farOutVec:Set(1e6, 1e6)

		if (not (self.taperStart and self.taperEnd)) then
-- equalize the widths at the endpoints, so they can be connected smoothly

			local tempW = mesh:Point(self.startPoint + 1).fWidth.value
			mesh:Point(self.startPoint).fWidth:SetValue(moho.frame, tempW)
			tempW = mesh:Point(mesh:CountPoints() - 2).fWidth.value
			mesh:Point(mesh:CountPoints() - 1).fWidth:SetValue(moho.frame, tempW)
		end

		mesh:AddPoint(farOutVec, -1, moho.frame)
		tempW = mesh:Point(mesh:CountPoints() - 3).fWidth.value
		mesh:Point(mesh:CountPoints() - 2).fWidth:SetValue(moho.frame, tempW)
		mesh:WeldPoints(mesh:CountPoints() - 2, mesh:CountPoints() - 3, moho.frame)
		tempW = mesh:Point(self.startPoint).fWidth.value
		mesh:Point(mesh:CountPoints() - 1).fWidth:SetValue(moho.frame, tempW)
		mesh:WeldPoints(mesh:CountPoints() - 1, self.startPoint, moho.frame)
	end

-- select the points in this new curve

	for i = self.startPoint, mesh:CountPoints() - 1 do
		mesh:Point(i).fSelected = true
	end
	if (self.weldToPt >= 0) then
		mesh:Point(self.weldToPt).fSelected = true
	end

-- create a shape based on this new curve

	if (FA_SharedUtils.autoFill) then
		local shapeID = moho:CreateShape(true)

		if (shapeID >= 0) then
			if (not FA_SharedUtils.autoOutline) then
				mesh:Shape(shapeID).fHasOutline = false
			end
		elseif (FA_SharedUtils.autoOutline) then
			moho:CreateShape(false)
		end
	elseif (FA_SharedUtils.autoOutline) then
		moho:CreateShape(false)
	end
	moho:NewKeyframe(CHANNEL_FILL)
	moho:NewKeyframe(CHANNEL_LINE)

	moho:AddPointKeyframe(moho.frame)
	mouseEvent.view:DrawMe()
	moho:NewKeyframe(CHANNEL_POINT)
	moho:NewKeyframe(CHANNEL_CURVE)
	moho:UpdateSelectedChannels()
end

function FA_Freehand:OnKeyDown(moho, keyEvent)

	if (self.mouseon and keyEvent.key == ' ') then
		local mesh= moho:Mesh()
		if (mesh == nil) then return end

		if (self.breakPoint < mesh:CountPoints()) then
			mesh:Point(self.breakPoint).fSelected= false
		end
		mesh:Point(mesh:CountPoints()-1).fSelected= false

		MOHO.DeleteSelectedPoints(mesh)
		self.breakPoint= self.breakPoint + 1
		keyEvent.view:DrawMe()
		return
	end
	FA_SharedUtils:KeyDown(moho, keyEvent)
end

function FA_Freehand:WeldEndpoint(moho, view, startID, endID)
	local mesh = moho:Mesh()
	if (mesh == nil) then
		mouseEvent.view:DrawMe()
		return false
	end

	local testVec1 = LM.Vector2:new_local()
	local testVec2 = LM.Vector2:new_local()

	testVec1:Set(mesh:Point(endID).fPos)
	local closestID= mesh:ClosestPoint(testVec1, endID, startID)

	if (closestID < 0) then
		return false
	end

	testVec2:Set(mesh:Point(closestID).fPos)

-- convert the two chosen points to pixel coordinates
-- this is to make sure that they are actually close enough in screen space

	local p1 = LM.Point:new_local()
	local p2 = LM.Point:new_local()
	local mat = LM.Matrix:new_local()
	moho.layer:GetFullTransform(moho.frame, mat, moho.document)
	mat:Transform(testVec1)
	mat:Transform(testVec2)
	view:Graphics():WorldToScreen(testVec1, p1)
	view:Graphics():WorldToScreen(testVec2, p2)
	p1.x = p1.x - p2.x
	p1.y = p1.y - p2.y

	if (((p1.x * p1.x) + (p1.y * p1.y)) < 100) then

-- found one close enough to weld

		if (mesh:WeldPoints(endID, closestID, moho.frame)) then
			mesh:Point(closestID).fSelected = true
			return true
		end
	end

	return false
end

-- **************************************************
-- Freehand dialog
-- **************************************************

local FA_FreehandDialog = {}

function FA_FreehandDialog:new()
	local dlg = LM.GUI.SimpleDialog("Freehand Options", FA_FreehandDialog)
	local l = dlg:GetLayout()

	l:PushH()
		l:PushV()
			l:AddChild(LM.GUI.StaticText("Pixel tolerance"), LM.GUI.ALIGN_LEFT)
			l:AddChild(LM.GUI.StaticText("Angle tolerance"), LM.GUI.ALIGN_LEFT)
		l:Pop()
		l:PushV()
			dlg.pixelText = LM.GUI.TextControl(0, "000", 0, LM.GUI.FIELD_UINT)
			l:AddChild(dlg.pixelText, LM.GUI.ALIGN_LEFT)
			dlg.angleText = LM.GUI.TextControl(0, "000", 0, LM.GUI.FIELD_UINT)
			l:AddChild(dlg.angleText, LM.GUI.ALIGN_LEFT)
		l:Pop()
	l:Pop()

	l:AddChild(LM.GUI.StaticText("Variable line width:"), LM.GUI.ALIGN_LEFT)
	l:Indent()

	dlg.lineWidthNone = LM.GUI.RadioButton(FA_SharedUtils.txNone, FA_Freehand.PRESSURE)
	l:AddChild(dlg.lineWidthNone, LM.GUI.ALIGN_LEFT)
	dlg.lineWidthPressure = LM.GUI.RadioButton("Use pen pressure", FA_Freehand.PRESSURE)
	l:AddChild(dlg.lineWidthPressure, LM.GUI.ALIGN_LEFT)
	dlg.lineWidthRandom = LM.GUI.RadioButton("Random", FA_Freehand.PRESSURE)
	l:AddChild(dlg.lineWidthRandom, LM.GUI.ALIGN_LEFT)
	dlg.lineWidthFixed = LM.GUI.RadioButton("Fixed", FA_Freehand.PRESSURE)
	l:AddChild(dlg.lineWidthFixed, LM.GUI.ALIGN_LEFT)

	l:PushH()
		l:Unindent()
		l:PushV()
			l:AddChild(LM.GUI.StaticText(FA_SharedUtils.txMinWidth), LM.GUI.ALIGN_LEFT)
			l:AddChild(LM.GUI.StaticText(FA_SharedUtils.txMaxWidth), LM.GUI.ALIGN_LEFT)
		l:Pop()

		l:PushV()
			dlg.minText = LM.GUI.TextControl(0, "00.000", 0, LM.GUI.FIELD_UFLOAT)
			dlg.minText:SetWheelInc(1)
			l:AddChild(dlg.minText, LM.GUI.ALIGN_LEFT)
			dlg.maxText = LM.GUI.TextControl(0, "00.000", 0, LM.GUI.FIELD_UFLOAT)
			dlg.maxText:SetWheelInc(1)
			l:AddChild(dlg.maxText, LM.GUI.ALIGN_LEFT)
		l:Pop()
	l:Pop()

	dlg.taperStartCheck = LM.GUI.CheckBox("Taper start")
	l:AddChild(dlg.taperStartCheck, LM.GUI.ALIGN_LEFT)

	dlg.taperEndCheck = LM.GUI.CheckBox("Taper end")
	l:AddChild(dlg.taperEndCheck, LM.GUI.ALIGN_LEFT)

	l:AddChild(LM.GUI.Button(FA_SharedUtils.txReset, FA_Freehand.RESET))
	return dlg
end

function FA_FreehandDialog:UpdateWidgets()
	self.pixelText:SetValue(FA_Freehand.pixelTolerance)
	self.angleText:SetValue(FA_Freehand.angleTolerance)
	if (FA_Freehand.widthMode == 1) then
		self.lineWidthPressure:SetValue(true)
	elseif (FA_Freehand.widthMode == 2) then
		self.lineWidthRandom:SetValue(true)
	elseif (FA_Freehand.widthMode == 3) then
		self.lineWidthFixed:SetValue(true)
	else
		self.lineWidthNone:SetValue(true)
	end
	self.minText:SetValue(FA_Freehand.minWidth)
	self.maxText:SetValue(FA_Freehand.maxWidth)
	self.minText:Enable(FA_Freehand.widthMode > 0)
	self.maxText:Enable(FA_Freehand.widthMode ~= 0 and FA_Freehand.widthMode ~= 3)
	self.taperStartCheck:SetValue(FA_Freehand.taperStart)
	self.taperEndCheck:SetValue(FA_Freehand.taperEnd)
end

function FA_FreehandDialog:OnValidate()
	local b = true
	if (not self:Validate(self.pixelText, 2, 1000)) then
		b = false
	end
	if (not self:Validate(self.angleText, 0, 30)) then
		b = false
	end
	if (not self:Validate(self.minText, 0, 32)) then
		b = false
	end
	if (not self:Validate(self.maxText, self.minText:FloatValue(), 32)) then
		b = false
	end
	return b
end

function FA_FreehandDialog:OnOK()
	FA_Freehand.pixelTolerance = self.pixelText:IntValue()
	FA_Freehand.angleTolerance = self.angleText:IntValue()

	if (self.lineWidthPressure:Value()) then
		FA_Freehand.widthMode = 1
	elseif (self.lineWidthRandom:Value()) then
		FA_Freehand.widthMode = 2
	elseif (self.lineWidthFixed:Value()) then
		FA_Freehand.widthMode = 3
	else
		FA_Freehand.widthMode = 0
	end
	FA_Freehand.minWidth = self.minText:FloatValue()
	FA_Freehand.maxWidth = self.maxText:FloatValue()
	FA_Freehand.taperStart = self.taperStartCheck:Value()
	FA_Freehand.taperEnd = self.taperEndCheck:Value()
end

function FA_FreehandDialog:HandleMessage(msg)
	if (msg == FA_Freehand.PRESSURE) then
		self.minText:Enable(not self.lineWidthNone:Value())
		self.maxText:Enable(not (self.lineWidthNone:Value() or self.lineWidthFixed:Value()))
	elseif (msg == FA_Freehand.RESET) then
		FA_Freehand.pixelTolerance = 10
		FA_Freehand.angleTolerance = 10
		FA_Freehand.widthMode = 1
		FA_Freehand.minWidth = 1
		FA_Freehand.maxWidth = 8
		FA_Freehand.taperStart = false
		FA_Freehand.taperEnd = false
		self:UpdateWidgets()
	end
end

-- **************************************************
-- Tool options - create and respond to tool's UI
-- **************************************************

FA_Freehand.AUTOCLOSE = MOHO.MSG_BASE
FA_Freehand.AUTOFILL = MOHO.MSG_BASE + 1
FA_Freehand.AUTOOUTLINE = MOHO.MSG_BASE + 2
FA_Freehand.AUTOWELD = MOHO.MSG_BASE + 3
FA_Freehand.PRESSURE = MOHO.MSG_BASE + 4
FA_Freehand.RESET = MOHO.MSG_BASE + 5
FA_Freehand.SHARPCORNERS = MOHO.MSG_BASE + 6

function FA_Freehand:DoLayout(moho, layout)
	self.autoCloseCheck = LM.GUI.CheckBox(FA_SharedUtils.txAutoClose, self.AUTOCLOSE)
	layout:AddChild(self.autoCloseCheck)

	self.autoFillCheck = LM.GUI.CheckBox(FA_SharedUtils.txAutoFill, self.AUTOFILL)
	layout:AddChild(self.autoFillCheck)

	self.autoOutlineCheck = LM.GUI.CheckBox(FA_SharedUtils.txAutoOutline, self.AUTOOUTLINE)
	layout:AddChild(self.autoOutlineCheck)

	self.dlog = FA_FreehandDialog:new()
	self.popup = LM.GUI.PopupDialog("Freehand Options")
	self.popup:SetDialog(self.dlog)
	layout:AddChild(self.popup)

	self.autoWeldCheck = LM.GUI.CheckBox(FA_SharedUtils.txAutoWeld, self.AUTOWELD)
	layout:AddChild(self.autoWeldCheck)

	self.sharpCornersCheck = LM.GUI.CheckBox(FA_SharedUtils.txSharpCorners, self.SHARPCORNERS)
	layout:AddChild(self.sharpCornersCheck)
end

function FA_Freehand:UpdateWidgets(moho)

	if (FA_SharedUtils.autoFill and not(FA_SharedUtils.autoClose)) then
		FA_SharedUtils.autoClose = true
	end
	self.autoCloseCheck:SetValue(FA_SharedUtils.autoClose)
	self.autoFillCheck:SetValue(FA_SharedUtils.autoFill)
	self.autoOutlineCheck:SetValue(FA_SharedUtils.autoOutline)
	self.autoWeldCheck:SetValue(FA_SharedUtils.autoWeld)
	self.sharpCornersCheck:SetValue(FA_SharedUtils.sharpCorners)
end

function FA_Freehand:HandleMessage(moho, view, msg)
	if (msg == self.AUTOCLOSE) then
		FA_SharedUtils.autoClose = self.autoCloseCheck:Value()
		if (FA_SharedUtils.autoFill and not(FA_SharedUtils.autoClose)) then
			FA_SharedUtils.autoFill = false
			self.autoFillCheck:SetValue(FA_SharedUtils.autoFill)
		end
	elseif (msg == self.AUTOFILL) then
		FA_SharedUtils.autoFill = self.autoFillCheck:Value()
		if (FA_SharedUtils.autoFill and not(FA_SharedUtils.autoClose)) then
			FA_SharedUtils.autoClose = true
			self.autoCloseCheck:SetValue(FA_SharedUtils.autoClose)
		end
	elseif (msg == self.AUTOOUTLINE) then
		FA_SharedUtils.autoOutline = self.autoOutlineCheck:Value()
	elseif (msg == self.AUTOWELD) then
		FA_SharedUtils.autoWeld = self.autoWeldCheck:Value()
	elseif (msg == self.SHARPCORNERS) then
		FA_SharedUtils.sharpCorners = self.sharpCornersCheck:Value()
	end
end
