-- **************************************************
-- Provide Moho with the name of this script object
-- **************************************************

ScriptName = "FA_Magnet"

-- **************************************************
-- General information about this script
-- **************************************************

FA_Magnet = {}

function FA_Magnet:Name()
	return "Magnet"
end

function FA_Magnet:Version()
	return "5.0"
end

function FA_Magnet:Description()
	return "Move points within a radius of influence (hold <ctrl> to constrain, <space> to disconnect mirror points)"
end

function FA_Magnet:Creator()
	return "Lost Marble, Modified by Fazek"
end

function FA_Magnet:UILabel()
	return("Magnet")
end

function FA_Magnet:LoadPrefs(prefs)
	self.magnetRadius = prefs:GetFloat("FA_Magnet.magnetRadius", 0.5)
end

function FA_Magnet:SavePrefs(prefs)
	prefs:SetFloat("FA_Magnet.magnetRadius", self.magnetRadius)
end

-- **************************************************
-- Recurring values
-- **************************************************

FA_Magnet.magnetRadius = 0.5
FA_Magnet.mirror = 0 -- 0: no mirror, 1: central, 2: axial mirroring
FA_Magnet.dragging = 0 -- 0: no drag, 1: drag a single point, 2: normal, 3,5: mirror point A, 4,6: mirror point B

FA_Magnet.dragVec1 = LM.Vector2:new_local()
FA_Magnet.dragVec2 = LM.Vector2:new_local()
FA_Magnet.mirrorA = LM.Vector2:new_local()
FA_Magnet.mirrorB = LM.Vector2:new_local()
FA_Magnet.mirrorPtA = -1
FA_Magnet.mirrorPtB = -1

FA_Magnet.minVec = LM.Vector2:new_local()
FA_Magnet.maxVec = LM.Vector2:new_local()

-- **************************************************
-- The guts of this script
-- **************************************************

function FA_Magnet:IsEnabled(moho)
	if (moho:CountPoints() > 0) then
		return true
	end
	return false
end

function FA_Magnet:OnMouseDown(moho, mouseEvent)

	self.dragging= 0

	local mesh = moho:Mesh()
	if (mesh == nil) then return end

	if (mouseEvent.shiftKey and not mouseEvent.altKey) then
		FA_SharedUtils.selectionMode= 0
		FA_SharedUtils:MouseDown(moho, mouseEvent)
	else
		local m = LM.Matrix:new_local()
		local screenPt= LM.Point:new_local()
		local v = LM.Vector2:new_local()

		if (self.mirror ~= 0) then
	--clicked on mirror handles?
			self:GetMirror(mesh)

			moho.layer:GetFullTransform(moho.frame, m, moho.document)
			v:Set(self.mirrorA)
			m:Transform(v)
			mouseEvent.view:Graphics():WorldToScreen(v, screenPt)

			local dX= mouseEvent.pt.x - screenPt.x
			local dY= mouseEvent.pt.y - screenPt.y

			if ((dX * dX) + (dY * dY) < 100) then
				self.dragging= 3

				if (self.mirrorPtA >= 0) then
					mesh:Point(self.mirrorPtA).fSelected= true
				end

			elseif (self.mirror == 2) then
				v:Set(self.mirrorB)

				m:Transform(v)
				mouseEvent.view:Graphics():WorldToScreen(v, screenPt)

				dX= mouseEvent.pt.x - screenPt.x
				dY= mouseEvent.pt.y - screenPt.y

				if ((dX * dX) + (dY * dY) < 100) then
					self.dragging= 4

					if (self.mirrorPtB >= 0) then
						mesh:Point(self.mirrorPtB).fSelected= true
					end
				end
			end
		end

		if (self.dragging == 0) then

			local ptID= mouseEvent.view:PickPoint(mouseEvent.pt)

			if (ptID >= 0) then
				local pt= mesh:Point(ptID)

				if (pt.fSelected) then
		--clicking on an existing selected point
					moho.document:PrepUndo(moho.layer)
					moho.document:SetDirty()
					mesh:PrepMovePoints()
					self.dragVec1:Set(mouseEvent.startVec)
					self.dragging= 1
					mouseEvent.view:DrawMe()
					return
				end

			else

		--clicked on an edge?

				local cveID, segID= -1, -1
				cveID, segID= mouseEvent.view:PickEdge(mouseEvent.pt, cveID,segID)

				if (cveID >= 0) then
					moho.document:PrepUndo(moho.layer)
					moho.document:SetDirty()

					local cve= mesh:Curve(cveID)
					mesh:AddPoint(mouseEvent.vec, cveID, segID, moho.frame)

					if (self.mirror ~= 0) then
		--'m' already filled
						local ax= mouseEvent.vec.x - self.mirrorA.x
						local ay= mouseEvent.vec.y - self.mirrorA.y

						if (self.mirror == 1) then
			--central mirroring
							ax= self.mirrorA.x - ax
							ay= self.mirrorA.y - ay

							v:Set(ax, ay)
							m:Transform(v)
							mouseEvent.view:Graphics():WorldToScreen(v, screenPt)
							cveID, segID= mouseEvent.view:PickEdge(screenPt, cveID,segID)
						else
			--axial mirroring
							local mx= self.mirrorB.x - self.mirrorA.x
							local my= self.mirrorB.y - self.mirrorA.y

							local q= (mx * mx) + (my * my)

							if (q > 0) then
								local x= ((ay * mx) - (ax * my)) / q
								local y= ((ax * mx) + (ay * my)) / q

								ax= self.mirrorA.x + (mx * y) + (my * x)
								ay= self.mirrorA.y + (my * y) - (mx * x)

								v:Set(ax, ay)
								m:Transform(v)
								mouseEvent.view:Graphics():WorldToScreen(v, screenPt)
								cveID, segID= mouseEvent.view:PickEdge(screenPt, cveID,segID)
							else
								cveID= -1
							end
						end

						if (cveID >= 0) then
		--mirror point
							v:Set(ax, ay)
							cve= mesh:Curve(cveID)
							mesh:AddPoint(v, cveID, segID, moho.frame)
						end
					end
					mesh:PrepMovePoints()
					self.dragVec1:Set(mouseEvent.startVec)
					self.dragging= 1
					mouseEvent.view:DrawMe()
					return
				end
			end

			if (moho:CountSelectedPoints() > 0) then
				FA_SharedUtils.selectionMode= 1
				FA_SharedUtils:MouseDown(moho, mouseEvent)
			else
				FA_SharedUtils.selectionMode= 0
				FA_SharedUtils:MouseDown(moho, mouseEvent)
			end
		else
--moving the mirror
			if (self.mirrorPtA >= 0) then
		--marking to move the points as well
				self.dragging= self.dragging + 2
				moho.document:PrepUndo(moho.layer)
				moho.document:SetDirty()
				mouseEvent.view:DrawMe()
			end
		end
	end
end

function FA_Magnet:OnMouseMoved(moho, mouseEvent)
	local mesh = moho:Mesh()
	if (mesh == nil) then return end

	if (self.dragging > 2) then

		if (self.dragging == 3 or self.dragging == 5) then
--drag mirror point A
			self.mirrorA:Set(mouseEvent.vec)

			if (mouseEvent.ctrlKey) then
				if (math.abs(self.mirrorA.x - self.mirrorB.x) > math.abs(self.mirrorA.y - self.mirrorB.y)) then
					self.mirrorA.y= self.mirrorB.y
				else
					self.mirrorA.x= self.mirrorB.x
				end
			end

			if (moho.gridOn) then
				moho:SnapToGrid(self.mirrorA)
			end

			if (self.mirrorPtA >= 0) then
				mesh:Point(self.mirrorPtA).fPos:Set(self.mirrorA)
				moho:AddPointKeyframe(moho.frame)
			end
		else
--drag mirror point B
			self.mirrorB:Set(mouseEvent.vec)

			if (mouseEvent.ctrlKey) then
				if (math.abs(self.mirrorA.x - self.mirrorB.x) > math.abs(self.mirrorA.y - self.mirrorB.y)) then
					self.mirrorB.y= self.mirrorA.y
				else
					self.mirrorB.x= self.mirrorA.x
				end
			end

			if (moho.gridOn) then
				moho:SnapToGrid(self.mirrorB)
			end

			if (self.mirrorPtB >= 0) then
				mesh:Point(self.mirrorPtB).fPos:Set(self.mirrorB)
				moho:AddPointKeyframe(moho.frame)
			end
		end
		mouseEvent.view:DrawMe()
		return
	end

	if (FA_SharedUtils.mouseMode ~= 0) then
		FA_SharedUtils:MouseMoved(moho, mouseEvent)

		if (FA_SharedUtils.mouseMode == 0) then
			moho.document:PrepUndo(moho.layer)
			moho.document:SetDirty()
			mesh:PrepMovePoints()
		else
--waiting for movement
			return
		end
	end
	self.dragging= 2
	self.dragVec1:Set(mouseEvent.vec)
	local startVec2= LM.Vector2:new_local()

	self:DragPoints(mesh, self.dragVec1, mouseEvent.startVec, false)
	self:GetMirror(mesh)

	if (self.mirror ~= 0) then
		local ax= mouseEvent.vec.x - self.mirrorA.x
		local ay= mouseEvent.vec.y - self.mirrorA.y
		local bx= mouseEvent.startVec.x - self.mirrorA.x
		local by= mouseEvent.startVec.y - self.mirrorA.y

		if (self.mirror == 1) then
--central mirroring
			ax= self.mirrorA.x - ax
			ay= self.mirrorA.y - ay
			bx= self.mirrorA.x - bx
			by= self.mirrorA.y - by

			self.dragVec2:Set(ax, ay)
			startVec2:Set(bx, by)
			self:DragPoints(mesh, self.dragVec2, startVec2, true)
		else
--axial mirroring
			local mx= self.mirrorB.x - self.mirrorA.x
			local my= self.mirrorB.y - self.mirrorA.y

			local q= (mx * mx) + (my * my)

			if (q > 0) then
				local x= ((ay * mx) - (ax * my)) / q
				local y= ((ax * mx) + (ay * my)) / q

				ax= self.mirrorA.x + (mx * y) + (my * x)
				ay= self.mirrorA.y + (my * y) - (mx * x)

				x= ((by * mx) - (bx * my)) / q
				y= ((bx * mx) + (by * my)) / q
				bx= self.mirrorA.x + (mx * y) + (my * x)
				by= self.mirrorA.y + (my * y) - (mx * x)

				self.dragVec2:Set(ax, ay)
				startVec2:Set(bx, by)
				self:DragPoints(mesh, self.dragVec2, startVec2, true)
			else
				self.dragVec2:Set(mouseEvent.vec)
			end
		end
	end
	moho:AddPointKeyframe(moho.frame)
	mouseEvent.view:DrawMe()
end

function FA_Magnet:OnMouseUp(moho, mouseEvent)

	if (self.dragging > 2 and self.dragging <= 4) then
		self.dragging= 0
		return
	end

	local mesh = moho:Mesh()
	if (mesh == nil) then
		return
	end

	if (FA_SharedUtils.mouseMode > 0) then
		FA_SharedUtils:MouseUp(moho, mouseEvent)
		return
	end

	self.dragging = 0
	moho:AddPointKeyframe(moho.frame)
	moho:NewKeyframe(CHANNEL_POINT)
	mouseEvent.view:DrawMe()
end

function FA_Magnet:OnKeyDown(moho, keyEvent)
	local mesh= moho:Mesh()
	if (mesh == nil) then return end

	if (keyEvent.key == ' ') then

		if (self.mirror ~= 0 and self.mirrorPtA >= 0) then
			self.mirrorPtA= -1
			self.mirrorPtB= -1
			keyEvent.view:DrawMe()
		end
	else
		FA_SharedUtils:KeyDown(moho, keyEvent)
	end
end

function FA_Magnet:DrawMe(moho, view)
	local mesh = moho:Mesh()

	if (mesh == nil) then return end

	if (self.dragging == 0 and self.mirror == 0) then
		return
	end

	local gfx = view:Graphics()
	local matrix = LM.Matrix:new_local()

	moho.layer:GetFullTransform(moho.frame, matrix, moho.document)
	gfx:Push()
	gfx:ApplyMatrix(matrix)

	gfx:SetColor(255, 0, 0)

	if (self.mirror == 1) then
--central mirror
		self:GetMirror(mesh)
		gfx:DrawLine(self.mirrorA.x - 0.05, self.mirrorA.y, self.mirrorA.x + 0.05, self.mirrorA.y)
		gfx:DrawLine(self.mirrorA.x, self.mirrorA.y - 0.05, self.mirrorA.x, self.mirrorA.y + 0.05)

		if (self.mirrorPtA < 0) then
			gfx:DrawFatMarker(self.mirrorA.x, self.mirrorA.y, 4)
		end

	elseif (self.mirror == 2) then
--axial mirror
		self:GetMirror(mesh)
		local aX,aY= self.mirrorA.x, self.mirrorA.y
		local bX,bY= (self.mirrorB.x - aX) * 0.5,(self.mirrorB.y - aY) * 0.5

		gfx:DrawLine(aX - bX, aY - bY, self.mirrorB.x + bX, self.mirrorB.y + bY)

		if (self.mirrorPtA < 0) then
			gfx:DrawFatMarker(self.mirrorA.x, self.mirrorA.y, 4)
			gfx:DrawFatMarker(self.mirrorB.x, self.mirrorB.y, 4)
		end
	end

	if (self.dragging == 1 or self.dragging == 2) then
		gfx:SetColor(255, 0, 0, 128)
		gfx:SetSmoothing(true)
		gfx:FillCircle(self.dragVec1, self.magnetRadius)

		if (self.mirror > 0 and self.dragging == 2) then
			gfx:SetColor(255, 0, 0, 64)
			gfx:FillCircle(self.dragVec2, self.magnetRadius)
		end
		gfx:SetSmoothing(false)
	end
	gfx:Pop()
end

function FA_Magnet:DragPoints(mesh, vec, startVec, usePos)

	local offset= vec - startVec

	for i= 0, mesh:CountPoints() - 1 do
		local pt= mesh:Point(i)

		if (pt.fSelected) then
			local v= pt.fTempPos - startVec
			local magInfluence= v:Mag() / self.magnetRadius

			if (magInfluence < 1.0) then
				magInfluence= 1.0 - (magInfluence * magInfluence)

				if (usePos) then
					pt.fPos= pt.fPos + offset * magInfluence
				else
					pt.fPos= pt.fTempPos + offset * magInfluence
				end

			elseif (not usePos) then
				pt.fPos= pt.fTempPos
			end
		end
	end
end

function FA_Magnet:GetMirror(mesh)

	if (self.mirrorPtA >= 0) then
		self.mirrorA:Set(mesh:Point(self.mirrorPtA).fPos)
		self.mirrorB:Set(mesh:Point(self.mirrorPtB).fPos)
	end
end

-- **************************************************
-- Tool options - create and respond to tool's UI
-- **************************************************

FA_Magnet.CHANGE = MOHO.MSG_BASE
FA_Magnet.DUMMY = MOHO.MSG_BASE + 1
FA_Magnet.RESET = MOHO.MSG_BASE + 2
FA_Magnet.SETMIRROR = MOHO.MSG_BASE + 3
FA_Magnet.SELECTITEM = MOHO.MSG_BASE + 10

function FA_Magnet:DoLayout(moho, layout)
	self.menu = LM.GUI.Menu("Select Group")

	self.popup = LM.GUI.PopupMenu(128, false)
	self.popup:SetMenu(self.menu)
	layout:AddChild(self.popup)

	layout:AddChild(LM.GUI.StaticText("Magnet radius"))
	self.radius = LM.GUI.TextControl(0, "00.0000", self.CHANGE, LM.GUI.FIELD_UFLOAT)
	self.radius:SetWheelInc(0.1)
	layout:AddChild(self.radius)

	self.resetBut= LM.GUI.Button(FA_SharedUtils.txCopy0, self.RESET)
	layout:AddChild(self.resetBut)

	self.mirrMenu= LM.GUI.Menu("Set mirror")
	self.mirrMenu:AddItem("Off", 0, self.SETMIRROR)
	self.mirrMenu:AddItem("Central", 0, self.SETMIRROR + 1)
	self.mirrMenu:AddItem("Axial", 0, self.SETMIRROR + 2)
	self.mirrMenu:AddItem("By selected points", 0, self.SETMIRROR + 3)
	self.mirrPopup= LM.GUI.PopupMenu(128, false)
	self.mirrPopup:SetMenu(self.mirrMenu)
	layout:AddChild(self.mirrPopup)

	self.mirrInfo= LM.GUI.DynamicText("",300)
	layout:AddChild(self.mirrInfo)

	if (self.mirror > 0) then

--checking mirror
		if (self.mirrorPtA >= 0) then
			local mesh= moho:Mesh()

			if (mesh == nil) then
				self.mirrorPtA= -1
				self.mirrorPtB= -1
			else
				local ptN= mesh:CountPoints()

				if (ptN <= self.mirrorPtA) then
					self.mirrorPtA= -1
					self.mirrorPtB= -1
				else
					local pt= mesh:Point(self.mirrorPtA).fPos

					if (pt.x ~= self.mirrorA.x or pt.y ~= self.mirrorA.y) then
						self.mirrorPtA= -1
						self.mirrorPtB= -1
					end
				end

				if (self.mirror == 2) then

					if (ptN <= self.mirrorPtB) then
						self.mirrorPtA= -1
						self.mirrorPtB= -1
					else
						local pt= mesh:Point(self.mirrorPtB).fPos

						if (pt.x ~= self.mirrorB.x or pt.y ~= self.mirrorB.y) then
							self.mirrorPtA= -1
							self.mirrorPtB= -1
						end
					end
				end
			end
		end
	end
end

function FA_Magnet:UpdateWidgets(moho)
	self.resetBut:Enable(moho.frame ~= 0)

	local mesh = moho:Mesh()
	if (mesh == nil) then return end

	MOHO.BuildGroupMenu(self.menu, mesh, self.SELECTITEM, self.DUMMY)
	self.radius:SetValue(self.magnetRadius)

	local mirrName= "no mirroring"

	if (self.mirror == 1) then
		mirrName= "Central mirroring"
	elseif (self.mirror == 2) then
		mirrName= "Axial mirroring"
	end
	self.mirrInfo:SetValue(mirrName)
end

function FA_Magnet:HandleMessage(moho, view, msg)
	local mesh = moho:Mesh()
	if (mesh == nil) then
		return
	end

	if (msg == self.RESET) then

		if (moho.frame > 0) then
			local mesh = moho:Mesh()
			if (mesh == nil) then return end

			moho.document:PrepUndo(moho.layer)
			moho.document:SetDirty()

			for i = 0, mesh:CountPoints() - 1 do
				local pt = mesh:Point(i)
				if (pt.fSelected) then
					pt:SetPos(pt.fAnimPos:GetValue(0), moho.frame)
				end
			end
			moho:NewKeyframe(CHANNEL_POINT)
		end

	elseif (msg == self.CHANGE) then
		self.magnetRadius = self.radius:FloatValue()
		if (self.magnetRadius < 0.001) then
			self.magnetRadius = 0.001
		end
	elseif (msg >= self.SELECTITEM) then
		FA_SharedUtils:SelectItem(moho, mesh, msg - self.SELECTITEM)
	elseif (msg >= self.SETMIRROR) then

		if (msg >= self.SETMIRROR + 3) then
--set mirror mode: central if one point, axial if two (or more...) selected

			self.mirror= 0

			for i= 0, mesh:CountPoints() - 1 do
				local pt= mesh:Point(i)

				if (pt.fSelected) then
					self.mirror= self.mirror + 1

					if (self.mirror == 2) then
				--axial
						self.mirrorPtB= i
						self.mirrorB:Set(pt.fPos)
						break
					else
						self.mirrorPtA= i
						self.mirrorA:Set(pt.fPos)
					end
				end
			end

			if (self.mirror ~= 0) then
				moho.document:PrepUndo(moho.layer)
				mesh:SelectNone()
			end
			moho:UpdateUI()
		else
			self.mirror= msg - self.SETMIRROR
			self.mirrorA:Set(0,0)
			self.mirrorB:Set(0,0.5)
			self.mirrorPtA= -1
			self.mirrorPtB= -1
			moho:UpdateUI()
		end
	end
end
