-- **************************************************
-- Provide Moho with the name of this script object
-- **************************************************

ScriptName = "FA_PaintBucket"

-- **************************************************
-- General information about this script
-- **************************************************

FA_PaintBucket = {}

function FA_PaintBucket:Name()
	return "Paint Bucket"
end

function FA_PaintBucket:Version()
	return "5.0"
end

function FA_PaintBucket:Description()
	return "Click to fill a closed region with color (click-drag to preview the fill)"
end

function FA_PaintBucket:Creator()
	return "Lost Marble, Modified by Fazek"
end

function FA_PaintBucket:UILabel()
	return("Paint Bucket")
end

-- **************************************************
-- Recurring values
-- **************************************************

FA_PaintBucket.dragging = false

-- **************************************************
-- The guts of this script
-- **************************************************

function FA_PaintBucket:IsEnabled(moho)
	if (moho.layer:CurrentAction() ~= "") then
		return false -- creating new objects in the middle of an action can lead to unexpected results
	end
	if (moho:CountEdges() > 0) then
		return true
	end
	return false
end

function FA_PaintBucket:OnMouseDown(moho, mouseEvent)
	self.dragging = true
	self:FloodSelect(moho, mouseEvent.view, mouseEvent.pt, 1)
	mouseEvent.view:DrawMe()
end

function FA_PaintBucket:OnMouseMoved(moho, mouseEvent)
	self:FloodSelect(moho, mouseEvent.view, mouseEvent.pt, 1)
	mouseEvent.view:DrawMe()
end

function FA_PaintBucket:OnMouseUp(moho, mouseEvent)

	if (self:FloodSelect(moho, mouseEvent.view, mouseEvent.pt, 1)) then
		local mesh = moho:Mesh()
		moho.document:PrepUndo(moho.layer)
		moho.document:SetDirty()
	-- only filled shapes

		local shapeID = moho:CreateShape(true)

		if (shapeID >= 0) then
			moho:NewKeyframe(CHANNEL_FILL)
			moho:NewKeyframe(CHANNEL_LINE)
			moho:UpdateSelectedChannels()
		else
	--try to fill with the shape with the other method
			mesh:SelectNone()

			if (FA_SharedUtils:FloodSelection(moho,mouseEvent)) then
				shapeID = moho:CreateShape(true)

				if (shapeID >= 0) then
					moho:NewKeyframe(CHANNEL_FILL)
					moho:NewKeyframe(CHANNEL_LINE)
					moho:UpdateSelectedChannels()
				end
			end
		end
	end

	self.dragging = false
	mouseEvent.view:DrawMe()
end

function FA_PaintBucket:OnKeyDown(moho, keyEvent)
	FA_SharedUtils:KeyDown(moho,keyEvent)
end

function FA_PaintBucket:DrawMe(moho, view)
	if (self.dragging) then
		local g = view:Graphics()
		local matrix = LM.Matrix:new_local()

		moho.layer:GetFullTransform(moho.frame, matrix, moho.document)
		g:Push()
		g:ApplyMatrix(matrix)

		view:DrawPreviewShape()

		g:Pop()
	end
end

function FA_PaintBucket:FloodSelect(moho, view, pt)
	view:FloodSelect(pt, 1)

	if (moho:CountSelectedEdges() > 0) then
--test for closed outlines
		local mesh= moho:Mesh()

		FA_SharedUtils:CollectSegments(mesh)
		local segList= FA_SharedUtils:CollectSegments(mesh)
		local tabN= table.getn(segList)

		if (tabN < 1) then
			return false
		end

		local cve, min, max, t, pt

	--if a segment has at least one active point except endpoints, extend to all points

		for i= 1, tabN do
			cve= mesh:Curve(segList[i][1])
			min= segList[i][2]
			max= segList[i][3]
			t= false

		--pass1: test

			if (max <= min) then
		--roundover
				for j= min + 1, cve:CountSegments() - 1 do

					if (cve:Point(j).fSelected) then
						t= true
						break
					end
				end

				if (not t) then

					for j= 0, max - 1 do

						if (cve:Point(j).fSelected) then
							t= true
							break
						end
					end
				end
			else
				for j= min + 1, max - 1 do

					if (cve:Point(j).fSelected) then
						t= true
						break
					end
				end
			end

		--pass2: set
			if (t) then

				if (max <= min) then

					for j= min, cve:CountSegments() - 1 do
						cve:Point(j).fSelected= true
					end
					min= 0
				end

				for j= min, max do
					cve:Point(j).fSelected= true
				end
			end
		end
		return true
	end
	return false
end
