-- **************************************************
-- Provide Moho with the name of this script object
-- **************************************************

ScriptName = "FA_ReparentBone"

-- **************************************************
-- General information about this script
-- **************************************************

FA_ReparentBone = {}

function FA_ReparentBone:Name()
	return "Reparent Bone"
end

function FA_ReparentBone:Version()
	return "5.0"
end

function FA_ReparentBone:Description()
	return "Attach bone(s) to a new parent (press <shift> to add selection, <space> to swap bone ends)"
end

function FA_ReparentBone:Creator()
	return "Lost Marble, Modified by Fazek"
end

function FA_ReparentBone:UILabel()
	return("Reparent Bone")
end

-- **************************************************
-- Recurring values
-- **************************************************

FA_ReparentBone.clickedID= -1
FA_ReparentBone.parentID= -1

-- **************************************************
-- The guts of this script
-- **************************************************

function FA_ReparentBone:IsEnabled(moho)
	if (moho.frame > 0) then
		return false
	end
	if (moho:CountBones() < 2) then
		return false
	end
	return true
end

function FA_ReparentBone:OnMouseDown(moho, mouseEvent)

	self.clickedID= -1
	self.parentID= -1

	if (mouseEvent.shiftKey and not mouseEvent.altKey) then
--force selection
		FA_SharedUtils.selectionMode= 0
		FA_SharedUtils:MouseDown(moho, mouseEvent)
	else
		local bnID= mouseEvent.view:PickBone(mouseEvent.pt, mouseEvent.vec, moho.layer, true)

		if (bnID >= 0) then

			if (moho:CountSelectedBones() < 1) then
--select a new bone
				local skel= moho:Skeleton()

				if (skel == nil) then return end

				moho.document:PrepUndo(moho.layer)

				for i= 0,skel:CountBones() - 1 do
					skel:Bone(i).fSelected= false
				end
				skel:Bone(bnID).fSelected= true
				FA_SharedUtils.mouseMode= 0
				mouseEvent.view:DrawMe()
			else
				FA_SharedUtils.mouseMode= -1
			end
		else
			FA_SharedUtils.selectionMode= 3
			FA_SharedUtils:MouseDown(moho, mouseEvent)

			if (FA_SharedUtils.mouseMode >= 0 and moho:CountSelectedBones() < 1) then
				FA_SharedUtils.selectionMode= 0
			end
		end
		self.clickedID= bnID
	end
end

function FA_ReparentBone:OnMouseMoved(moho, mouseEvent)

	if (FA_SharedUtils.mouseMode ~= 0) then
		FA_SharedUtils:MouseMoved(moho, mouseEvent)

		if (FA_SharedUtils.mouseMode == 0) then

			if (self.clickedID >= 0) then
--bone-to-parent drag
				if (mouseEvent.view:PickBone(mouseEvent.pt, mouseEvent.vec, moho.layer, false) == self.clickedID) then
		--fallback: waiting for the first valid parent
					FA_SharedUtils.mouseMode= -1
					return
				end
				local skel= moho:Skeleton()
				if (skel == nil) then return end

				moho.document:PrepUndo(moho.layer)
				moho.document:SetDirty()

				for i= 0, skel:CountBones() - 1 do
					skel:Bone(i).fSelected= false
				end
				skel:Bone(self.clickedID).fSelected= true
			end
		else
--waiting for movement
			return
		end
	end
	local parentID= mouseEvent.view:PickBone(mouseEvent.pt, mouseEvent.vec, moho.layer, false)
	self.parentID= parentID
	self:ReparentBones(moho,parentID)
	mouseEvent.view:DrawMe()
	moho:UpdateSelectedChannels()
end

function FA_ReparentBone:OnMouseUp(moho, mouseEvent)
	self.parentID= -1

	if (FA_SharedUtils.mouseMode >= 3) then
		FA_SharedUtils:MouseUp(moho,mouseEvent)
		return
	end

	self.statusInfo:SetValue("")

	if (FA_SharedUtils.mouseMode ~= 0) then
--just one click
		if (self.clickedID >= 0) then
			moho.document:PrepUndo(moho.layer)

			if (moho:CountSelectedBones() > 0) then
--on parent
				moho.document:SetDirty()
				self:ReparentBones(moho,self.clickedID)
			else
--select a new bone
				local skel= moho:Skeleton()

				if (skel == nil) then return end

				for i= 0, skel:CountBones() - 1 do
					skel:Bone(i).fSelected= false
				end
				skel:Bone(self.clickedID).fSelected= true
			end
		else
			FA_SharedUtils:MouseUp(moho,mouseEvent)
			return
		end
	end
	moho.layer:UpdateCurFrame()
	moho:NewKeyframe(CHANNEL_BONE)
	mouseEvent.view:DrawMe()
	moho:UpdateSelectedChannels()
end

function FA_ReparentBone:OnKeyDown(moho, keyEvent)

	if (keyEvent.key == ' ') then

		if (self.parentID >= 0) then return end

		if (moho:CountSelectedBones() < 1) then return end

		local skel = moho:Skeleton()
		if (skel == nil) then return end

		moho.document:PrepUndo(moho.layer)
		moho.document:SetDirty()

		local invMatrix = LM.Matrix:new_local()
		local v1 = LM.Vector2:new_local()
		local v2 = LM.Vector2:new_local()
		local v3 = LM.Vector2:new_local()

--swap the bone's two ends

		for i = 0, skel:CountBones() - 1 do
			local bone = skel:Bone(i)

			if (bone.fSelected) then
				v1:Set(bone.fLength, 0)
				v2:Set(0, 0)
				bone.fRestMatrix:Transform(v1)
				bone.fRestMatrix:Transform(v2)

				if (bone.fParent >= 0) then
					local parent= skel:Bone(bone.fParent)

					invMatrix:Set(parent.fRestMatrix)
					invMatrix:Invert()
					invMatrix:Transform(v1)
					invMatrix:Transform(v2)

					if (parent.fSelected) then
						v3:Set(parent.fLength, 0)
						v1 = v3 - v1
						v2 = v3 - v2
					end
				end

				v2 = v2 - v1
				local angle = math.atan2(v2.y, v2.x)

				while angle > 2 * math.pi do
					angle = angle - 2 * math.pi
				end
				while angle < 0 do
					angle = angle + 2 * math.pi
				end
				bone.fAnimPos:SetValue(0, v1)
				bone.fAnimAngle:SetValue(0, angle)

			elseif (bone.fParent >= 0) then
				local parent= skel:Bone(bone.fParent)

				if (parent.fSelected) then
					v1:Set(0, 0)
					v2:Set(bone.fLength, 0)
					bone.fRestMatrix:Transform(v1)
					bone.fRestMatrix:Transform(v2)

					invMatrix:Set(parent.fRestMatrix)
					invMatrix:Invert()
					invMatrix:Transform(v1)
					invMatrix:Transform(v2)

					v3:Set(parent.fLength, 0)
					v2 = v1 - v2
					v1 = v3 - v1

					local angle = math.atan2(v2.y, v2.x)

					while angle > 2 * math.pi do
						angle = angle - 2 * math.pi
					end
					while angle < 0 do
						angle = angle + 2 * math.pi
					end
					bone.fAnimPos:SetValue(0, v1)
					bone.fAnimAngle:SetValue(0, angle)
				end
			end
		end
		moho.layer:UpdateCurFrame()
		moho:NewKeyframe(CHANNEL_BONE)
		keyEvent.view:DrawMe()
		moho:UpdateSelectedChannels()
	else
		FA_SharedUtils:KeyDown(moho, keyEvent)
	end
end

function FA_ReparentBone:DrawMe(moho, view)
	local skel = moho:Skeleton()
	if (skel == nil) then return end

	local gfx = view:Graphics()
	local matrix = LM.Matrix:new_local()
	local bonePt = LM.Vector2:new_local()
	local parentPt = LM.Vector2:new_local()
	local arrowPt1 = LM.Vector2:new_local()
	local arrowPt2 = LM.Vector2:new_local()
	local mag, f

	moho.layer:GetFullTransform(moho.frame, matrix, moho.document)
	gfx:Push()
	gfx:ApplyMatrix(matrix)

	gfx:SetColor(255, 0, 0)
	gfx:SetSmoothing(true)

	for i = 0, skel:CountBones() - 1 do
		local bone = skel:Bone(i)

		if (bone.fParent >= 0) then
			local parentBone = skel:Bone(bone.fParent)

			bonePt:Set(bone.fLength / 2, 0)
			bone.fRestMatrix:Transform(bonePt)

			parentPt:Set(parentBone.fLength / 2, 0)
			parentBone.fRestMatrix:Transform(parentPt)

			local v = parentPt - bonePt
			mag = v:Mag()

			if (mag > 0.2) then
				f = 0.0125
				mag = mag * 0.75
			else
				f = mag / 16
				mag= mag - 0.05
			end
			arrowPt1:Set(mag, -f)
			arrowPt2:Set(mag, f)
			f = math.atan2(v.y, v.x)
			arrowPt1:Rotate(f)
			arrowPt2:Rotate(f)
			arrowPt1.x = arrowPt1.x + bonePt.x
			arrowPt1.y = arrowPt1.y + bonePt.y
			arrowPt2.x = arrowPt2.x + bonePt.x
			arrowPt2.y = arrowPt2.y + bonePt.y

			gfx:DrawLine(bonePt.x, bonePt.y, parentPt.x, parentPt.y)
			gfx:BeginShape()
			gfx:AddLine(parentPt, arrowPt1)
			gfx:AddLine(arrowPt1, arrowPt2)
			gfx:AddLine(arrowPt2, parentPt)
			gfx:EndShape()
		end
	end

	if (self.parentID >= 0) then
		local parentBone= skel:Bone(self.parentID)

		parentPt:Set(parentBone.fLength / 2, 0)
		parentBone.fRestMatrix:Transform(parentPt)

		gfx:DrawLine(parentPt.x - 0.05, parentPt.y, parentPt.x + 0.05, parentPt.y)
		gfx:DrawLine(parentPt.x, parentPt.y - 0.05, parentPt.x, parentPt.y + 0.05)
	end
	gfx:Pop()
end

function FA_ReparentBone:ReparentBones(moho, parentID)
	local skel = moho:Skeleton()
	if (skel == nil) then return end

	local invMatrix = LM.Matrix:new_local()
	local v1 = LM.Vector2:new_local()
	local v2 = LM.Vector2:new_local()
	local pID

	for i = 0, skel:CountBones() - 1 do
		local bone = skel:Bone(i)

		if (bone.fSelected) then

			if (i == parentID) then
				pID= -1
			else
				pID= parentID
			end

			if (bone.fParent ~= pID) then

				if (not skel:IsBoneChild(i, pID)) then
	-- new parent chosen
					v1:Set(0, 0)
					v2:Set(bone.fLength, 0)
					bone.fRestMatrix:Transform(v1)
					bone.fRestMatrix:Transform(v2)

					if (pID >= 0) then
						invMatrix:Set(skel:Bone(pID).fRestMatrix)
						invMatrix:Invert()
						invMatrix:Transform(v1)
						invMatrix:Transform(v2)
					end
					v2 = v2 - v1
					local angle = math.atan2(v2.y, v2.x)

					while angle > 2 * math.pi do
						angle = angle - 2 * math.pi
					end
					while angle < 0 do
						angle = angle + 2 * math.pi
					end
					bone.fAnimPos:SetValue(0, v1)
					bone.fAnimAngle:SetValue(0, angle)
					bone.fParent = pID

					if (pID < 0) then
						self.statusInfo:SetValue("(Itself: top-level bone)")
					else
						self.statusInfo:SetValue("(Parent bone)")
					end
				else
					self.statusInfo:SetValue("(Child bone: not possible)")
				end

			elseif (pID < 0) then
				self.statusInfo:SetValue("(Itself: top-level bone)")
			end
		end
	end
end

-- **************************************************
-- Tool options - create and respond to tool's UI
-- **************************************************

FA_ReparentBone.CHANGE = MOHO.MSG_BASE
FA_ReparentBone.DUMMY = MOHO.MSG_BASE + 1
FA_ReparentBone.SELECTBONE = MOHO.MSG_BASE + 2

function FA_ReparentBone:DoLayout(moho, layout)
	self.menu = LM.GUI.Menu("Select Bone")

	self.popup = LM.GUI.PopupMenu(128, false)
	self.popup:SetMenu(self.menu)
	layout:AddChild(self.popup)

	self.statusInfo = LM.GUI.DynamicText("",512)
	layout:AddChild(self.statusInfo)
end

function FA_ReparentBone:UpdateWidgets(moho)
	local skel = moho:Skeleton()
	if (skel == nil) then
		return
	end

	MOHO.BuildBoneMenu(self.menu, skel, self.SELECTBONE, self.DUMMY)
end

function FA_ReparentBone:HandleMessage(moho, view, msg)
	local skel = moho:Skeleton()
	if (skel == nil) then
		return
	end

	if (msg >= self.SELECTBONE) then
		FA_SharedUtils:SelectBone(moho, skel, msg - self.SELECTBONE, false)
	end
end
