-- **************************************************
-- Provide Moho with the name of this script object
-- **************************************************

ScriptName = "FA_RotatePoints"

-- **************************************************
-- General information about this script
-- **************************************************

FA_RotatePoints = {}

function FA_RotatePoints:Name()
	return "Rotate Points/Bones"
end

function FA_RotatePoints:Version()
	return "5.0"
end

function FA_RotatePoints:Description()
	return "Rotate selected points or bones (<ctrl> to constrain, <shift> to add selection/move with center, <alt> to scale, <space> to reset center)"
end

function FA_RotatePoints:Creator()
	return "Lost Marble, Modified by Fazek"
end

function FA_RotatePoints:UILabel()
	return("Rotate Points/Bones")
end

function FA_RotatePoints:LoadPrefs(prefs)
	FA_SharedUtils:Load_Prefs(prefs)
end

function FA_RotatePoints:SavePrefs(prefs)
	FA_SharedUtils:Save_Prefs(prefs)
end

-- **************************************************
-- Recurring values
-- **************************************************

FA_RotatePoints.dragging = false
FA_RotatePoints.dragCenter = 0 --0: off, 1:drag, 2:drag and move
FA_RotatePoints.startAngle = 0
FA_RotatePoints.lastScale = 1
FA_RotatePoints.lastVec = LM.Vector2:new_local()
FA_RotatePoints.centerVec = LM.Vector2:new_local()
FA_RotatePoints.centerAdd = LM.Vector2:new_local()
FA_RotatePoints.centerBackup = LM.Vector2:new_local()
FA_RotatePoints.addValid= false
FA_RotatePoints.addAdjust= false
FA_RotatePoints.boneMode= false

-- **************************************************
-- The guts of this script
-- **************************************************

function FA_RotatePoints:IsEnabled(moho)
	if (moho:CountPoints() > 0 or moho:CountBones() > 0) then
		return true
	end
	return false
end

function FA_RotatePoints:OnMouseDown(moho, mouseEvent)

	if (moho.layer:LayerType() == MOHO.LT_BONE) then
		self.boneMode= true

		if (moho:CountSelectedBones() < 2) then
			self.centerBackup:Set(0,0)
		else
			self.centerBackup:Set(FA_SharedUtils:SelectedBonesCenter(moho,FA_SharedUtils.centerGravity))
		end
	else
		self.boneMode= false

		if (moho:CountSelectedPoints() < 2) then
			self.centerBackup:Set(0,0)
		else
			self.centerBackup:Set(FA_SharedUtils:SelectedCenter(moho,FA_SharedUtils.centerGravity))
		end
	end

	if (not self.addValid) then
		self.centerAdd:Set(0,0)
		self.addValid= true
	end

	local m = LM.Matrix:new_local()
	local screenPt = LM.Point:new_local()

	moho.layer:GetFullTransform(moho.frame, m, moho.document)
	local v= self.centerBackup + self.centerAdd

	m:Transform(v)
	mouseEvent.view:Graphics():WorldToScreen(v,screenPt)

	local dX= mouseEvent.pt.x - screenPt.x
	local dY= mouseEvent.pt.y - screenPt.y

	if ((dX * dX) + (dY * dY) < 100) then

		if (self.boneMode) then
--bones
			if (mouseEvent.shiftKey and not mouseEvent.altKey and moho:CountSelectedBones() > 0) then
				self.dragCenter= 2
				moho.document:PrepUndo(moho.layer)
				moho.document:SetDirty()
				FA_SharedUtils:PrepMoveBones(moho)
				moho:UpdateBonePointSelection()
				mouseEvent.view:DrawMe()
				moho:UpdateSelectedChannels()
			else
				self.dragCenter= 1
			end
		else
--points
			if (mouseEvent.shiftKey and not mouseEvent.altKey and moho:CountSelectedPoints() > 0) then
				self.dragCenter= 2
				moho.document:PrepUndo(moho.layer)
				moho.document:SetDirty()

				moho:Mesh():PrepMovePoints()
				mouseEvent.view:DrawMe()
			else
				self.dragCenter= 1
			end
		end
		self.lastVec:Set(mouseEvent.vec)
		self.centerVec:Set(self.centerBackup)
		self.angle:SetValue(0)

	elseif (mouseEvent.shiftKey and not mouseEvent.altKey) then
--force selection
		self.dragCenter= 0
		FA_SharedUtils.selectionMode= 0
		FA_SharedUtils:MouseDown(moho,mouseEvent)
	else
		self.dragCenter= 0
		self.addAdjust= true

		if (self.boneMode) then
			FA_SharedUtils.selectionMode= 3

			FA_SharedUtils:MouseDown(moho, mouseEvent)

			if (FA_SharedUtils.mouseMode >= 0 and moho:CountSelectedBones() < 1) then
				FA_SharedUtils.selectionMode= 0
			end
		else
			FA_SharedUtils.selectionMode= 1
			FA_SharedUtils:MouseDown(moho, mouseEvent)

			if (FA_SharedUtils.mouseMode >= 0 and moho:CountSelectedPoints() < 1) then
				FA_SharedUtils.selectionMode= 0
			end
		end
		self.angle:SetValue(0)
	end
	self.lastScale = 1
	self.dragging = true
end

function FA_RotatePoints:OnMouseMoved(moho, mouseEvent)

	if (self.dragCenter > 0) then
		local vec= LM.Vector2:new_local()
		vec:Set(mouseEvent.vec)

		if (moho.gridOn) then
			moho:SnapToGrid(vec)
		end
		self.centerAdd= self.centerAdd + vec - self.lastVec
		self.lastVec:Set(vec)
		vec= vec - mouseEvent.startVec

		if (self.dragCenter > 1) then
--move points/bones
			if (self.boneMode) then
		--bones
				FA_SharedUtils:TranslateBones(moho,vec)
				moho.layer:UpdateCurFrame()
			else
		--points
				local mesh = moho:Mesh()
				if (mesh == nil) then return end

				mesh:TranslatePoints(vec)
				moho:AddPointKeyframe(moho.frame)
			end
		end
		mouseEvent.view:DrawMe()
		return
	end

	if (FA_SharedUtils.mouseMode ~= 0) then
		FA_SharedUtils:MouseMoved(moho,mouseEvent)

		if (FA_SharedUtils.mouseMode == 0) then

			if (self.boneMode) then
		--bones
				moho.document:PrepUndo(moho.layer)
				moho.document:SetDirty()

				FA_SharedUtils:PrepMoveBones(moho)
			else
		--points
				local mesh = moho:Mesh()
				if (mesh == nil) then return end

				moho.document:PrepUndo(moho.layer)
				moho.document:SetDirty()
				mesh:PrepMovePoints()
			end
			self.centerVec:Set(self.centerBackup)
			self.lastVec:Set(mouseEvent.vec)
			self.startAngle = 0
		else
--waiting for movement
			return
		end
	end

	local angle = self.startAngle
	local centerVec= self.centerVec + self.centerAdd
	local v1 = self.lastVec - centerVec
	local v2 = mouseEvent.vec - centerVec

	v2:Rotate(-math.atan2(v1.y, v1.x))
	angle = angle + math.atan2(v2.y, v2.x)
	self.startAngle = angle

	if (mouseEvent.ctrlKey) then
		angle = angle / (math.pi / 4)
		angle = (math.pi / 4) * LM.Round(angle)
	end

	self.angle:SetValue(math.floor(math.deg(angle) * 100) / 100)

	if (self.boneMode) then
--bones
		FA_SharedUtils:RotateBones(moho,angle,centerVec)
		moho.layer:UpdateCurFrame()
	else
--points
		local mesh = moho:Mesh()
		if (mesh == nil) then return end

		if (mouseEvent.altKey) then
			local v3= mouseEvent.startVec - centerVec
			local l1= v3:Mag()
			local l2= v2:Mag()

			if (l1 * 1000 <= l2) then
				self.lastScale= 0.001
			else
				self.lastScale= l2 / l1
			end
		end

		mesh:RotatePoints(angle, centerVec)

		if (self.lastScale ~= 1) then
--scale by hand
			for i= 0,mesh:CountPoints() - 1 do
				local pt= mesh:Point(i)

				if (pt.fSelected) then
					v3= ((pt.fPos - centerVec) * self.lastScale) + centerVec
					pt.fPos:Set(v3)
				end
			end
		end
		moho:AddPointKeyframe(moho.frame)
	end
	mouseEvent.view:DrawMe()
	self.lastVec:Set(mouseEvent.vec)
end

function FA_RotatePoints:OnMouseUp(moho, mouseEvent)
	self.dragging = false

	if (self.dragCenter > 0) then
		self.angle:SetValue(0)

		if (self.dragCenter > 1) then
			self.addAdjust= true
			self:AdjustAdd(moho)
		end
		self.dragCenter= 0
		return
	end
	self.addAdjust= true

	if (FA_SharedUtils.mouseMode > 0) then
		FA_SharedUtils:MouseUp(moho, mouseEvent)
		self:AdjustAdd(moho)
		return
	end

	if (self.boneMode) then
		moho.layer:UpdateCurFrame()
		moho:NewKeyframe(CHANNEL_BONE)
	else
		moho:AddPointKeyframe(moho.frame)
		moho:NewKeyframe(CHANNEL_POINT)
	end
	mouseEvent.view:DrawMe()
	self:AdjustAdd(moho)
end

function FA_RotatePoints:OnKeyDown(moho, keyEvent)

	if (keyEvent.key == ' ') then

		if (self.dragging) then return end

		if (self.addValid) then
			self.addValid= false
			keyEvent.view:DrawMe()
		end
	else
		FA_SharedUtils:KeyDown(moho, keyEvent)
	end
end

function FA_RotatePoints:DrawMe(moho, view)
	local centerVec = LM.Vector2:new_local()

	if (self.dragging) then
		centerVec:Set(self.centerVec)
	elseif (moho.layer:LayerType() == MOHO.LT_BONE) then

		if (moho:CountSelectedBones() < 2) then
			centerVec:Set(0,0)
		else
			centerVec:Set(FA_SharedUtils:SelectedBonesCenter(moho,FA_SharedUtils.centerGravity))
		end
	else
		if (moho:CountSelectedPoints() < 2) then
			centerVec:Set(0,0)
		else
			centerVec:Set(FA_SharedUtils:SelectedCenter(moho,FA_SharedUtils.centerGravity))
		end
	end

	if (self.addValid) then
		self:AdjustAdd(moho)

		if (centerVec.x ~= self.centerBackup.x or centerVec.y ~= self.centerBackup.y) then
			self.addValid= false
		else
			centerVec= centerVec + self.centerAdd
		end
	end
	local gfx = view:Graphics()
	local matrix = LM.Matrix:new_local()

	moho.layer:GetFullTransform(moho.frame, matrix, moho.document)
	gfx:Push()
	gfx:ApplyMatrix(matrix)

	gfx:SetColor(255, 0, 0)
	gfx:DrawLine(centerVec.x - 0.05, centerVec.y, centerVec.x + 0.05, centerVec.y)
	gfx:DrawLine(centerVec.x, centerVec.y - 0.05, centerVec.x, centerVec.y + 0.05)
	gfx:DrawFatMarker(centerVec.x, centerVec.y, 4)
	gfx:Pop()
end

function FA_RotatePoints:AdjustAdd(moho)

	if (not self.addAdjust) then return end

	self.addAdjust= false

	if (not self.addValid) then return end

--adjust centerAdd to remain in place...

	self.centerAdd= self.centerAdd + self.centerBackup

	if (moho.layer:LayerType() == MOHO.LT_BONE) then
--bones
		if (moho:CountSelectedBones() < 2) then
			self.centerBackup:Set(0,0)
		else
			self.centerBackup:Set(FA_SharedUtils:SelectedBonesCenter(moho,FA_SharedUtils.centerGravity))
		end
	else
--points
		if (moho:CountSelectedPoints() < 2) then
			self.centerBackup:Set(0,0)
		else
			self.centerBackup:Set(FA_SharedUtils:SelectedCenter(moho,FA_SharedUtils.centerGravity))
		end
	end
	self.centerAdd= self.centerAdd - self.centerBackup
end

-- **************************************************
-- Tool options - create and respond to tool's UI
-- **************************************************

FA_RotatePoints.DUMMY = MOHO.MSG_BASE
FA_RotatePoints.MODIFY = MOHO.MSG_BASE + 1
FA_RotatePoints.RESET = MOHO.MSG_BASE + 2
FA_RotatePoints.CENTER = MOHO.MSG_BASE + 3
FA_RotatePoints.LASSO = MOHO.MSG_BASE + 4
FA_RotatePoints.SELECTITEM = MOHO.MSG_BASE + 5

function FA_RotatePoints:DoLayout(moho, layout)
	self.groupMenu = LM.GUI.Menu(FA_SharedUtils.txSelectGroup)

	self.groupPopup = LM.GUI.PopupMenu(128, false)
	self.groupPopup:SetMenu(self.groupMenu)
	layout:AddChild(self.groupPopup)

	layout:AddChild(LM.GUI.StaticText(FA_SharedUtils.txAngle))

	self.angle = LM.GUI.TextControl(0, "000.0000", 0, LM.GUI.FIELD_FLOAT)
	self.angle:SetWheelInc(5)
	self.angle:SetValue(0)
	layout:AddChild(self.angle)

	self.modifyBut= LM.GUI.Button(FA_SharedUtils.txRotate, self.MODIFY)
	layout:AddChild(self.modifyBut)
	self.resetBut= LM.GUI.Button(FA_SharedUtils.txCopy0, self.RESET)
	layout:AddChild(self.resetBut)

	self.centerCheck = LM.GUI.CheckBox(FA_SharedUtils.txCenterGravity, self.CENTER)
	layout:AddChild(self.centerCheck)

	self.lassoCheck = LM.GUI.CheckBox(FA_SharedUtils.txLasso, self.LASSO)
	layout:AddChild(self.lassoCheck)
end

function FA_RotatePoints:UpdateWidgets(moho)
	self.centerCheck:SetValue(FA_SharedUtils.centerGravity)
	self.lassoCheck:SetValue(FA_SharedUtils.lassoMode)
	self.resetBut:Enable(moho.frame ~= 0)

	local mesh = moho:Mesh()

	if (mesh == nil) then
		self.groupMenu:RemoveAllItems()
		self.groupPopup:Enable(false)

		self.modifyBut:Enable(moho:CountSelectedBones() > 0)
	else
		MOHO.BuildGroupMenu(self.groupMenu, mesh, self.SELECTITEM, self.DUMMY)
		self.groupPopup:Enable(true)

		self.modifyBut:Enable(moho:CountSelectedPoints() > 0)
	end
end

function FA_RotatePoints:HandleMessage(moho, view, msg)

	if (msg == self.MODIFY) then
		moho.document:PrepUndo(moho.layer)
		moho.document:SetDirty()
		self.addAdjust= true

		local centerVec= LM.Vector2:new_local()
		local angle = math.rad(self.angle:FloatValue())

		if (moho.layer:LayerType() == MOHO.LT_BONE) then
--bones
			if (moho:CountSelectedBones() < 2) then
				centerVec:Set(0,0)
			else
				centerVec:Set(FA_SharedUtils:SelectedBonesCenter(moho,FA_SharedUtils.centerGravity))
			end

			if (self.addValid) then
				centerVec= centerVec + self.centerAdd
			end
			FA_SharedUtils:PrepMoveBones(moho)
			FA_SharedUtils:RotateBones(moho, angle, centerVec)
			moho.layer:UpdateCurFrame()
			moho:NewKeyframe(CHANNEL_BONE)
		else
--points
			local mesh = moho:Mesh()
			if (mesh == nil) then return end

			if (moho:CountSelectedPoints() < 2) then
				centerVec:Set(0,0)
			else
				centerVec:Set(FA_SharedUtils:SelectedCenter(moho,FA_SharedUtils.centerGravity))
			end

			if (self.addValid) then
				centerVec= centerVec + self.centerAdd
			end

			mesh:PrepMovePoints()
			mesh:RotatePoints(angle, centerVec)
			moho:AddPointKeyframe(moho.frame)
			moho:NewKeyframe(CHANNEL_POINT)
		end

	elseif (msg == self.RESET) then

		if (moho.layer:LayerType() == MOHO.LT_BONE) then
--bones
			local skel = moho:Skeleton()
			if (skel == nil) then return end

			if (moho.frame > 0) then
				moho.document:PrepUndo(moho.layer)
				moho.document:SetDirty()
				self.addAdjust= true

				for i = 0, skel:CountBones() - 1 do
					local bone = skel:Bone(i)

					if (bone.fSelected) then
						bone.fAnimAngle:SetValue(moho.frame, bone.fAnimAngle:GetValue(0))
						bone.fAnimPos:SetValue(moho.frame, bone.fAnimPos:GetValue(0))
					elseif (bone.fParent >= 0) then

						if (skel:Bone(bone.fParent).fSelected) then
							bone.fAnimAngle:SetValue(moho.frame, bone.fAnimAngle:GetValue(0))
							bone.fAnimPos:SetValue(moho.frame, bone.fAnimPos:GetValue(0))
						end
					end
				end
				moho.layer:UpdateCurFrame()
				moho:NewKeyframe(CHANNEL_BONE)
			end
		else
--points
			local mesh = moho:Mesh()
			if (mesh == nil) then return end

			if (moho.frame > 0) then
				moho.document:PrepUndo(moho.layer)
				moho.document:SetDirty()
				self.addAdjust= true

				for i = 0, mesh:CountPoints() - 1 do
					local pt = mesh:Point(i)
					if (pt.fSelected) then
						pt:SetPos(pt.fAnimPos:GetValue(0), moho.frame)
					end
				end
				moho:NewKeyframe(CHANNEL_POINT)
			end
		end
		self.angle:SetValue(0)
	elseif (msg >= self.SELECTITEM) then

		local mesh = moho:Mesh()
		if (mesh == nil) then return end

		self.angle:SetValue(0)
		FA_SharedUtils:SelectItem(moho, mesh, msg - self.SELECTITEM)
	elseif (msg == self.CENTER) then
		FA_SharedUtils.centerGravity = self.centerCheck:Value()
	elseif (msg == self.LASSO) then
		FA_SharedUtils.lassoMode = self.lassoCheck:Value()
	end
end
