-- **************************************************
-- Provide Moho with the name of this script object
-- **************************************************

ScriptName = "FA_ScalePoints"

-- **************************************************
-- General information about this script
-- **************************************************

FA_ScalePoints = {}

function FA_ScalePoints:Name()
	return "Scale Points"
end

function FA_ScalePoints:Version()
	return "5.0"
end

function FA_ScalePoints:Description()
	return "Scale selected points (<ctrl> to preserve volume, <shift> to add selection, <alt> to rotate)"
end

function FA_ScalePoints:Creator()
	return "Lost Marble, Modified by Fazek"
end

function FA_ScalePoints:UILabel()
	return("Scale Points")
end

-- **************************************************
-- Recurring values
-- **************************************************

FA_ScalePoints.dragging= false
FA_ScalePoints.scalingDir = -1
FA_ScalePoints.startVec= LM.Vector2:new_local()
FA_ScalePoints.centerVec= LM.Vector2:new_local()
FA_ScalePoints.lastAngle = 0
FA_ScalePoints.baseAngle = nil
FA_ScalePoints.lastTranslate = nil
FA_ScalePoints.lastPtX= 0
FA_ScalePoints.lastPtY= 0
FA_ScalePoints.startBoxW = 0
FA_ScalePoints.startBoxH = 0
FA_ScalePoints.boxW = 0
FA_ScalePoints.boxH = 0

-- **************************************************
-- The guts of this script
-- **************************************************

function FA_ScalePoints:IsEnabled(moho)
	if (moho:CountPoints() > 0) then
		return true
	end
	return false
end

function FA_ScalePoints:OnMouseDown(moho, mouseEvent)

	self.dragging= false

	local mesh = moho:Mesh()
	if (mesh == nil) then return end

	if (mouseEvent.shiftKey and not mouseEvent.altKey) then
		FA_SharedUtils.selectionMode= 0
		FA_SharedUtils:MouseDown(moho,mouseEvent)
	else
		self.scalingDir= -1

		if (moho:CountSelectedPoints() > 1) then
			local minVec,maxVec= FA_SharedUtils:SelectedBounds(moho,FA_SharedUtils.centerGravity)
			self.startVec:Set((minVec + maxVec) / 2)
			self.centerVec:Set(self.startVec)

-- First step - figure out which scaling handle is being dragged

			local boxW = maxVec.x - minVec.x
			local boxH = maxVec.y - minVec.y
			local dx = mouseEvent.vec.x - self.centerVec.x
			local dy = mouseEvent.vec.y - self.centerVec.y
			dx = math.abs(dx)
			dy = math.abs(dy)

			if ((dx > 0.4 * boxW) and (dx < 0.6 * boxW)) then

				if ((dy > 0.4 * boxH) and (dy < 0.6 * boxH)) then
					self.scalingDir = 2
				elseif (dy < 0.1 * boxH) then
					self.scalingDir = 0
				end
			elseif (dx < 0.1 * boxW) then

				if ((dy > 0.4 * boxH) and (dy < 0.6 * boxH)) then
					self.scalingDir = 1
				elseif (dy < 0.1 * boxH) then
					self.scalingDir = 3
					self.lastTranslate= nil
					self.lastPtX= mouseEvent.pt.x
					self.lastPtY= mouseEvent.pt.y
				end
			end
			self.lastAngle= 0
			self.baseAngle= nil
			self.startBoxW= boxW
			self.startBoxH= boxH
			self.boxW= boxW
			self.boxH= boxH

			if (self.scalingDir >= 0) then
				FA_SharedUtils.mouseMode= 0
				self.dragging= true
				moho.document:PrepUndo(moho.layer)
				moho.document:SetDirty()
				mesh:PrepMovePoints()

				mouseEvent.view:DrawMe()
			else
				self.scalingDir= 2
				FA_SharedUtils.selectionMode= 1
				FA_SharedUtils:MouseDown(moho, mouseEvent)
			end
		else
			FA_SharedUtils.selectionMode= 0
			FA_SharedUtils:MouseDown(moho, mouseEvent)
		end
	end
end

function FA_ScalePoints:OnMouseMoved(moho, mouseEvent)

	local mesh = moho:Mesh()
	if (mesh == nil) then return end

	if (FA_SharedUtils.mouseMode ~= 0) then
		FA_SharedUtils:MouseMoved(moho,mouseEvent)

		if (FA_SharedUtils.mouseMode ~= 0) then
			return
		end
		self.dragging= true
		moho.document:PrepUndo(moho.layer)
		moho.document:SetDirty()
		mesh:PrepMovePoints()
	end

	if (self.scalingDir < 0) then return end

	if (self.scalingDir == 3) then
--translate/rotate only

		local angle

		if (mouseEvent.altKey) then
			local v2 = mouseEvent.vec - self.centerVec
			angle= 0

			if (not self.baseAngle) then
				local dX= mouseEvent.pt.x - self.lastPtX
				local dY= mouseEvent.pt.y - self.lastPtY

				if ((dX * dX) + (dY * dY) < 100) then return end

				self.baseAngle= math.atan2(v2.y,v2.x) - self.lastAngle

			elseif (v2:Mag() > 0) then
				angle= math.atan2(v2.y,v2.x) - self.baseAngle
			end
			self.lastAngle= angle
		else
			local vec = LM.Vector2:new_local()

			vec:Set(mouseEvent.vec)

			if (moho.gridOn) then
				moho:SnapToGrid(vec)
			end
			self.lastTranslate= vec - mouseEvent.startVec
			self.lastPtX= mouseEvent.pt.x
			self.lastPtY= mouseEvent.pt.y
			self.baseAngle= nil
			self.centerVec= self.startVec + self.lastTranslate
			angle= self.lastAngle
		end

		if (self.lastTranslate) then
			mesh:TranslatePoints(self.lastTranslate)

			if (angle ~= 0) then

				for i= 0,mesh:CountPoints() - 1 do
					local pt= mesh:Point(i)

					if (pt.fSelected) then
						local vec= pt.fPos - self.centerVec
						vec:Rotate(angle)
						pt.fPos.x= vec.x + self.centerVec.x
						pt.fPos.y= vec.y + self.centerVec.y
					end
				end
			end
		elseif (angle ~= 0) then
			mesh:RotatePoints(angle, self.centerVec)
		end
	else
--scale/scale with rotate
		local v1 = mouseEvent.startVec - self.centerVec
		local v2 = mouseEvent.vec - self.centerVec

		if (mouseEvent.altKey) then
			v2:Rotate(-math.atan2(v1.y, v1.x))

			local angle= math.atan2(v2.y, v2.x)
			self.lastAngle= angle
			v2 = mouseEvent.vec - self.centerVec
		end

		if (self.lastAngle ~= 0) then
			v2:Rotate(-self.lastAngle)
		end

		local scaleX, scaleY

		if (self.scalingDir == 0) then
			scaleX= v2.x / v1.x

			if (mouseEvent.ctrlKey) then
				scaleY = 1 / scaleX
			else
				scaleY = 1
			end
		elseif (self.scalingDir == 1) then
			scaleY= v2.y / v1.y

			if (mouseEvent.ctrlKey) then
				scaleX = 1 / scaleY
			else
				scaleX = 1
			end
		else
			scaleX = v2:Mag() / v1:Mag()
			scaleY = scaleX
		end

		mesh:ScalePoints(scaleX, scaleY, self.centerVec)

		if (self.lastAngle ~= 0) then
--rotate by hand
			local x,y

			v1:Set(1,0)
			v1:Rotate(self.lastAngle)
			v2:Set(0,1)
			v2:Rotate(self.lastAngle)

			for i= 0,mesh:CountPoints() - 1 do
				local pt= mesh:Point(i)

				if (pt.fSelected) then
					x= pt.fPos.x - self.centerVec.x
					y= pt.fPos.y - self.centerVec.y
					pt.fPos.x= x*v1.x + y*v2.x + self.centerVec.x
					pt.fPos.y= x*v1.y + y*v2.y + self.centerVec.y
				end
			end			
		end
		self.scaleX:SetValue(scaleX)
		self.scaleY:SetValue(scaleY)
		self.boxW= self.startBoxW * scaleX
		self.boxH= self.startBoxH * scaleY
	end
	moho:AddPointKeyframe(moho.frame)
	mouseEvent.view:DrawMe()
end

function FA_ScalePoints:OnMouseUp(moho, mouseEvent)

	if (FA_SharedUtils.mouseMode > 0) then
		FA_SharedUtils:MouseUp(moho, mouseEvent)
		return
	end

	local mesh = moho:Mesh()
	if (mesh == nil) then return end

	self.dragging= false

	moho:AddPointKeyframe(moho.frame)
	moho:NewKeyframe(CHANNEL_POINT)

	mouseEvent.view:DrawMe()
end

function FA_ScalePoints:OnKeyDown(moho, keyEvent)
	FA_SharedUtils:KeyDown(moho, keyEvent)
end

function FA_ScalePoints:DrawMe(moho, view)
	local mesh = moho:Mesh()
	if (mesh == nil) then return end

	local vecX = LM.Vector2:new_local()
	local vecY = LM.Vector2:new_local()
	local aX,aY

	if (self.dragging) then
		vecX:Set(self.boxW,0)
		vecY:Set(0,self.boxH)

		if (self.lastAngle ~= 0) then
			vecX:Rotate(self.lastAngle)
			vecY:Rotate(self.lastAngle)
		end
		aX= self.centerVec.x - ((vecX.x + vecY.x) / 2)
		aY= self.centerVec.y - ((vecX.y + vecY.y) / 2)

	elseif (moho:CountSelectedPoints() > 1) then
		vecX,vecY = FA_SharedUtils:SelectedBounds(moho,FA_SharedUtils.centerGravity)
		aX= vecX.x
		aY= vecX.y
		vecX.x = vecY.x - vecX.x
		vecY.y = vecY.y - vecX.y
		vecX.y = 0
		vecY.x = 0
	else
		return
	end
	local bX= aX + vecX.x
	local bY= aY + vecX.y
	local cX= bX + vecY.x
	local cY= bY + vecY.y
	local dX= aX + vecY.x
	local dY= aY + vecY.y

	vecX.x= vecX.x / 2
	vecX.y= vecX.y / 2
	vecY.x= vecY.x / 2
	vecY.y= vecY.y / 2

	local gfx = view:Graphics()
	local markerR = 4
	local matrix = LM.Matrix:new_local()

	moho.layer:GetFullTransform(moho.frame, matrix, moho.document)
	gfx:Push()
	gfx:ApplyMatrix(matrix)

	gfx:SetColor(255, 0, 0)
	gfx:SetSmoothing(true)

	gfx:DrawLine(aX,aY,bX,bY)
	gfx:DrawLine(aX,aY,dX,dY)
	gfx:DrawLine(bX,bY,cX,cY)
	gfx:DrawLine(dX,dY,cX,cY)

	gfx:SetSmoothing(false)
	gfx:DrawFatMarker(aX, aY, markerR)
	gfx:DrawFatMarker(bX, bY, markerR)
	gfx:DrawFatMarker(cX, cY, markerR)
	gfx:DrawFatMarker(dX, dY, markerR)
	gfx:DrawFatMarker(aX + vecX.x, aY + vecX.y, markerR)
	gfx:DrawFatMarker(aX + vecY.x, aY + vecY.y, markerR)
	gfx:DrawFatMarker(bX + vecY.x, bY + vecY.y, markerR)
	gfx:DrawFatMarker(dX + vecX.x, dY + vecX.y, markerR)
	gfx:DrawFatMarker((aX + cX) / 2, (aY + cY) / 2, markerR)
	gfx:Pop()
end

-- **************************************************
-- Tool options - create and respond to tool's UI
-- **************************************************

FA_ScalePoints.DUMMY = MOHO.MSG_BASE
FA_ScalePoints.MODIFY = MOHO.MSG_BASE + 1
FA_ScalePoints.RESET = MOHO.MSG_BASE + 2
FA_ScalePoints.CENTER = MOHO.MSG_BASE + 3
FA_ScalePoints.LASSO = MOHO.MSG_BASE + 4
FA_ScalePoints.SELECTITEM = MOHO.MSG_BASE + 5

function FA_ScalePoints:DoLayout(moho, layout)
	self.menu = LM.GUI.Menu(FA_SharedUtils.txSelectGroup)

	self.popup = LM.GUI.PopupMenu(128, false)
	self.popup:SetMenu(self.menu)
	layout:AddChild(self.popup)

	layout:AddChild(LM.GUI.StaticText(FA_SharedUtils.txScale))

	layout:AddChild(LM.GUI.StaticText(FA_SharedUtils.txX))
	self.scaleX = LM.GUI.TextControl(0, "00.0000", 0, LM.GUI.FIELD_FLOAT)
	self.scaleX:SetValue(1)
	layout:AddChild(self.scaleX)

	layout:AddChild(LM.GUI.StaticText(FA_SharedUtils.txY))
	self.scaleY = LM.GUI.TextControl(0, "00.0000", 0, LM.GUI.FIELD_FLOAT)
	self.scaleY:SetValue(1)
	layout:AddChild(self.scaleY)

	self.modifyBut= LM.GUI.Button(FA_SharedUtils.txScale, self.MODIFY)
	layout:AddChild(self.modifyBut)
	self.resetBut= LM.GUI.Button(FA_SharedUtils.txCopy0, self.RESET)
	layout:AddChild(self.resetBut)

	self.centerCheck = LM.GUI.CheckBox(FA_SharedUtils.txCenterGravity, self.CENTER)
	layout:AddChild(self.centerCheck)

	self.lassoCheck = LM.GUI.CheckBox(FA_SharedUtils.txLasso, self.LASSO)
	layout:AddChild(self.lassoCheck)
end

function FA_ScalePoints:UpdateWidgets(moho)
	local mesh = moho:Mesh()
	if (mesh == nil) then return end

	MOHO.BuildGroupMenu(self.menu, mesh, self.SELECTITEM, self.DUMMY)
	self.modifyBut:Enable(moho:CountSelectedPoints() > 1)
	self.resetBut:Enable(moho.frame ~= 0)
	self.centerCheck:SetValue(FA_SharedUtils.centerGravity)
	self.lassoCheck:SetValue(FA_SharedUtils.lassoMode)
end

function FA_ScalePoints:HandleMessage(moho, view, msg)
	local mesh = moho:Mesh()
	if (mesh == nil) then
		return
	end

	if (msg == self.MODIFY) then

		if (moho:CountSelectedPoints() > 1) then
			moho.document:PrepUndo(moho.layer)
			moho.document:SetDirty()

			local centerVec = mesh:SelectedCenter()
			local scaleX = self.scaleX:FloatValue()
			local scaleY = self.scaleY:FloatValue()

			mesh:PrepMovePoints()
			mesh:ScalePoints(scaleX, scaleY, centerVec)
			moho:AddPointKeyframe(moho.frame)
			moho:NewKeyframe(CHANNEL_POINT)
		end
	elseif (msg == self.RESET) then
		if (moho.frame > 0) then
			moho.document:PrepUndo(moho.layer)
			moho.document:SetDirty()

			for i = 0, mesh:CountPoints() - 1 do
				local pt = mesh:Point(i)
				if (pt.fSelected) then
					pt:SetPos(pt.fAnimPos:GetValue(0), moho.frame)
				end
			end
		end
		moho:NewKeyframe(CHANNEL_POINT)
		self.scaleX:SetValue(1)
		self.scaleY:SetValue(1)
	elseif (msg >= self.SELECTITEM) then
		self.scaleX:SetValue(1)
		self.scaleY:SetValue(1)
		FA_SharedUtils:SelectItem(moho, mesh, msg - self.SELECTITEM)
	elseif (msg == self.CENTER) then
		FA_SharedUtils.centerGravity = self.centerCheck:Value()
	elseif (msg == self.LASSO) then
		FA_SharedUtils.lassoMode = self.lassoCheck:Value()
	end
end
