-- **************************************************
-- Provide Moho with the name of this script object
-- **************************************************

ScriptName = "FA_SelectPoints"

-- **************************************************
-- General information about this script
-- **************************************************

FA_SelectPoints = {}

function FA_SelectPoints:Name()
	return "Select Points/Bones"
end

function FA_SelectPoints:Version()
	return "5.0"
end

function FA_SelectPoints:Description()
	return "Select/deselect points, bones and shapes (press <ctrl> to override lasso mode/select full curves)"
end

function FA_SelectPoints:Creator()
	return "Lost Marble, Modified by Fazek"
end

function FA_SelectPoints:UILabel()
	return("Select Points/Bones")
end

function FA_SelectPoints:LoadPrefs(prefs)
	FA_SharedUtils:Load_Prefs(prefs)
end

function FA_SelectPoints:SavePrefs(prefs)
	FA_SharedUtils:Save_Prefs(prefs)
end

-- **************************************************
-- The guts of this script
-- **************************************************

function FA_SelectPoints:IsEnabled(moho)
	local ltype= moho.layer:LayerType()

	if (ltype == MOHO.LT_VECTOR or ltype == MOHO.LT_BONE) then
		return true
	end
	return false
end

function FA_SelectPoints:OnMouseDown(moho, mouseEvent)
	FA_SharedUtils.selectionMode= 0
	FA_SharedUtils:MouseDown(moho, mouseEvent)
end

function FA_SelectPoints:OnMouseMoved(moho, mouseEvent)
	FA_SharedUtils:MouseMoved(moho, mouseEvent)
end

function FA_SelectPoints:OnMouseUp(moho, mouseEvent)
	FA_SharedUtils:MouseUp(moho, mouseEvent)
end

function FA_SelectPoints:OnKeyDown(moho,keyEvent)
	FA_SharedUtils:KeyDown(moho,keyEvent)
end

-- **************************************************
-- Tool options - create and respond to tool's UI
-- **************************************************

FA_SelectPoints.SETGROUP = MOHO.MSG_BASE
FA_SelectPoints.DELGROUP = MOHO.MSG_BASE + 1
FA_SelectPoints.LASSO = MOHO.MSG_BASE + 2
FA_SelectPoints.DUMMY = MOHO.MSG_BASE + 3
FA_SelectPoints.SELECTITEM = MOHO.MSG_BASE + 4

function FA_SelectPoints:DoLayout(moho, layout)
	self.groupMenu= LM.GUI.Menu(FA_SharedUtils.txSelectGroup)
	self.groupPopup= LM.GUI.PopupMenu(128,false)
	self.groupPopup:SetMenu(self.groupMenu)
	layout:AddChild(self.groupPopup)

	self.groupName = LM.GUI.TextControl(0,"Room For a Big Long Name",0,LM.GUI.FIELD_TEXT)
	self.groupName:SetValue("")
	layout:AddChild(self.groupName)

	self.btSet= LM.GUI.Button(FA_SharedUtils.txSetGroup,self.SETGROUP)
	layout:AddChild(self.btSet)

	self.btDelete= LM.GUI.Button(FA_SharedUtils.txDeleteGroup,self.DELGROUP)
	layout:AddChild(self.btDelete)

	self.lassoCheck = LM.GUI.CheckBox(FA_SharedUtils.txLasso,self.LASSO)
	layout:AddChild(self.lassoCheck)
end

function FA_SelectPoints:UpdateWidgets(moho)
	local mesh= moho:Mesh()

	if (mesh == nil) then
		self.groupMenu:RemoveAllItems()
		self.groupPopup:Enable(false)

		self.btSet:Enable(false)
		self.btDelete:Enable(false)
	else
		MOHO.BuildGroupMenu(self.groupMenu,mesh,self.SELECTITEM,self.DUMMY)
		self.groupPopup:Enable(true)

		self.btSet:Enable(true)
		self.btDelete:Enable(true)
	end
	self.lassoCheck:SetValue(FA_SharedUtils.lassoMode)
end

function FA_SelectPoints:HandleMessage(moho,view,msg)

	if (msg == self.LASSO) then
		FA_SharedUtils.lassoMode = self.lassoCheck:Value()
		return
	end
	local mesh= moho:Mesh()

	if (mesh == nil) then return end

	if (msg == self.SETGROUP or msg == self.DELGROUP) then
		local name = self.groupName:Value()

		if (string.len(name) == 0) then
			LM.GUI.Alert(LM.GUI.ALERT_INFO,"To create or to modify a group, you must first name it.")
		elseif (msg == self.DELGROUP or moho:CountSelectedPoints() < 1) then
	--delete even if no points selected
			moho.document:PrepUndo(moho.layer)
			moho.document:SetDirty()
			mesh:DeleteGroup(name)
			self.groupName:SetValue("")
			moho:UpdateUI()
		else
	--set
			moho.document:PrepUndo(moho.layer)
			moho.document:SetDirty()
			mesh:AddGroup(name)
			moho:UpdateUI()
		end
	elseif (msg >= self.SELECTITEM) then
		FA_SharedUtils:SelectItem(moho, mesh, msg - self.SELECTITEM)
	end
end
