-- **************************************************
-- Provide Moho with the name of this script object
-- **************************************************

ScriptName = "FA_TranslatePoints"

-- **************************************************
-- General information about this script
-- **************************************************

FA_TranslatePoints = {}

function FA_TranslatePoints:Name()
	return "Translate Points/Bones"
end

function FA_TranslatePoints:Version()
	return "5.0"
end

function FA_TranslatePoints:Description()
	return "Move selected points or bones (hold <ctrl> to constrain, <shift> to add selection, <space> to weld, hold <alt> to disable auto-welding)"
end

function FA_TranslatePoints:Creator()
	return "Lost Marble, Modified by Fazek"
end

function FA_TranslatePoints:UILabel()
	return("Translate Points/Bones")
end

function FA_TranslatePoints:LoadPrefs(prefs)
	FA_SharedUtils:Load_Prefs(prefs)
	self.dragClosest = prefs:GetBool("FA_TranslatePoints.dragClosest", false)
end

function FA_TranslatePoints:SavePrefs(prefs)
	FA_SharedUtils:Save_Prefs(prefs)
	prefs:SetBool("FA_TranslatePoints.dragClosest", self.dragClosest)
end

-- **************************************************
-- Recurring values
-- **************************************************

FA_TranslatePoints.dragging = false
FA_TranslatePoints.numSel = 0
FA_TranslatePoints.selID = -1
FA_TranslatePoints.dragClosest = false
FA_TranslatePoints.boneMode = false

-- **************************************************
-- The guts of this script
-- **************************************************

function FA_TranslatePoints:IsEnabled(moho)
	if (moho:CountPoints() > 0 or moho:CountBones() > 0) then
		return true
	end
	return false
end

function FA_TranslatePoints:OnMouseDown(moho, mouseEvent)

	if (moho.layer:LayerType() == MOHO.LT_BONE) then
		self.boneMode= true
	else
		self.boneMode= false
	end

	if (mouseEvent.shiftKey and not mouseEvent.altKey) then
--force selection
		FA_SharedUtils.selectionMode= 0
		FA_SharedUtils:MouseDown(moho, mouseEvent)
	else
		if (self.boneMode) then
--bones
			FA_SharedUtils.selectionMode= 3
			FA_SharedUtils:MouseDown(moho, mouseEvent)

			if (FA_SharedUtils.mouseMode >= 0 and not self.dragClosest and moho:CountSelectedBones() < 1) then
				FA_SharedUtils.selectionMode= 0
			end
		else
--points
			FA_SharedUtils.selectionMode= 1
			FA_SharedUtils:MouseDown(moho, mouseEvent)

			if (FA_SharedUtils.mouseMode >= 0 and not self.dragClosest and moho:CountSelectedPoints() < 1) then
				FA_SharedUtils.selectionMode= 0
			end
		end
		self.textX:SetValue(0)
		self.textY:SetValue(0)
	end
	self.dragging= true
	self.selID= -1
end

function FA_TranslatePoints:OnMouseMoved(moho, mouseEvent)

	if (FA_SharedUtils.mouseMode ~= 0) then
		FA_SharedUtils:MouseMoved(moho, mouseEvent)

		if (FA_SharedUtils.mouseMode == 0) then

			if (self.boneMode) then
		--bones
				moho.document:PrepUndo(moho.layer)
				moho.document:SetDirty()

				FA_SharedUtils:PrepMoveBones(moho)
			else
		--points
				local mesh = moho:Mesh()
				if (mesh == nil) then return end

				moho.document:PrepUndo(moho.layer)
				moho.document:SetDirty()
				self.numSel = moho:CountSelectedPoints()

				if (self.numSel < 2) then

					if (self.dragClosest) then
--closest mode
						self.selID= mesh:ClosestPoint(mouseEvent.startVec)

						if (self.selID >= 0) then
							self.numSel = 1
							mesh:SelectNone()
							mesh:Point(self.selID).fSelected = true
							moho:UpdateSelectedChannels()
						end
					else
--single point mode
						self.selID= -1

						for i= 0, mesh:CountPoints() - 1 do

							if (mesh:Point(i).fSelected) then
								self.selID= i
								break
							end
						end
					end
				else
					self.selID= mesh:ClosestPoint(mouseEvent.startVec)
				end
				mesh:PrepMovePoints()
			end
		else
--waiting for movement
			return
		end
	end

	local curVec = mouseEvent.vec - mouseEvent.startVec

	if (mouseEvent.ctrlKey) then
		if (math.abs(mouseEvent.vec.x - mouseEvent.startVec.x) > math.abs(mouseEvent.vec.y - mouseEvent.startVec.y)) then
			curVec.y = 0
		else
			curVec.x = 0
		end
	end

	if (self.boneMode) then
--bones
		if (moho.gridOn) then
			moho:SnapToGrid(curVec)
		end
		FA_SharedUtils:TranslateBones(moho,curVec)
		moho.layer:UpdateCurFrame()
	else
--points
		local mesh = moho:Mesh()
		if (mesh == nil) then return end

		if (self.selID > -1) then
			local pt = mesh:Point(self.selID)

			if (moho.gridOn) then
				curVec:Set(pt.fTempPos + curVec)
				moho:SnapToGrid(curVec)
				curVec= curVec - pt.fTempPos
			end

			if (self.numSel < 2) then
-- just working with 1 point
				pt.fPos:Set(pt.fTempPos + curVec)
			else
-- move multiple points
				mesh:TranslatePoints(curVec)
			end

		elseif (self.numSel < 2) then
			return
		else
-- move multiple points
			if (moho.gridOn) then
				moho:SnapToGrid(curVec)
			end
			mesh:TranslatePoints(curVec)
		end
		moho:AddPointKeyframe(moho.frame)
	end
	self.textX:SetValue(curVec.x)
	self.textY:SetValue(curVec.y)
	mouseEvent.view:DrawMe()
end

function FA_TranslatePoints:OnMouseUp(moho, mouseEvent)

	if (FA_SharedUtils.mouseMode > 0) then
		FA_SharedUtils:MouseUp(moho, mouseEvent)
		return
	end

	if (self.boneMode) then
		moho.layer:UpdateCurFrame()
		moho:NewKeyframe(CHANNEL_BONE)
	else
		local mesh = moho:Mesh()
		if (mesh == nil) then return end

-- if we're just moving a single point, then try to auto-weld it

		if (self.selID >= 0 and self.numSel == 1 and FA_SharedUtils.autoWeld and not(mouseEvent.altKey)) then

			if (mesh:Point(self.selID):IsEndpoint()) then
				FA_SharedUtils:TryWeldPoints(moho,mouseEvent.view,true,self.selID)
			end
		end
		moho:AddPointKeyframe(moho.frame)
		moho:NewKeyframe(CHANNEL_POINT)
	end
	self.selID = -1
	self.dragging = false
	mouseEvent.view:DrawMe()
end

function FA_TranslatePoints:OnKeyDown(moho, keyEvent)
	local mesh = moho:Mesh()
	if (mesh == nil) then return end

	if (keyEvent.key == ' ') then
		local doit = false

		if (self.dragging) then
			if (self.numSel == 1) then
				doit = true
			end
		elseif (moho:CountSelectedPoints() == 1) then
			doit = true
		end

		if (doit) then
-- try welding the selected point to the closest other point

			for i = 0, mesh:CountPoints() - 1 do

				if (mesh:Point(i).fSelected) then
					FA_SharedUtils:TryWeldPoints(moho,keyEvent.view,self.dragging,i)
					self.selID= -1
					break
				end
			end
			keyEvent.view:DrawMe()
		end

	elseif (keyEvent.ctrlKey) then
		local inc = 1

		if (keyEvent.shiftKey) then
			inc = 10
		end
		local m = LM.Matrix:new_local()
		moho.layer:GetFullTransform(moho.frame, m, moho.document)

		if (keyEvent.keyCode == LM.GUI.KEY_UP) then
			FA_SharedUtils:TranslateStep(moho,mesh,keyEvent.view,0,-inc)
		elseif (keyEvent.keyCode == LM.GUI.KEY_DOWN) then
			FA_SharedUtils:TranslateStep(moho,mesh,keyEvent.view,0,inc)
		elseif (keyEvent.keyCode == LM.GUI.KEY_LEFT) then
			FA_SharedUtils:TranslateStep(moho,mesh,keyEvent.view,-inc,0)
		elseif (keyEvent.keyCode == LM.GUI.KEY_RIGHT) then
			FA_SharedUtils:TranslateStep(moho,mesh,keyEvent.view,inc,0)
		else
			FA_SharedUtils:KeyDown(moho, keyEvent)
		end
	else
		FA_SharedUtils:KeyDown(moho, keyEvent)
	end
end

function FA_TranslatePoints:DrawMe(moho, view)
	FA_SharedUtils:Draw_Me(moho, view, -1)
end

-- **************************************************
-- Tool options - create and respond to tool's UI
-- **************************************************

FA_TranslatePoints.DUMMY = MOHO.MSG_BASE
FA_TranslatePoints.MODIFY = MOHO.MSG_BASE + 1
FA_TranslatePoints.RESET = MOHO.MSG_BASE + 2
FA_TranslatePoints.AUTOWELD = MOHO.MSG_BASE + 3
FA_TranslatePoints.AUTOFILL = MOHO.MSG_BASE + 4
FA_TranslatePoints.LASSO = MOHO.MSG_BASE + 5
FA_TranslatePoints.CLOSEST = MOHO.MSG_BASE + 6
FA_TranslatePoints.SELECTITEM = MOHO.MSG_BASE + 7

function FA_TranslatePoints:DoLayout(moho, layout)
	self.groupMenu = LM.GUI.Menu(FA_SharedUtils.txSelectGroup)

	self.groupPopup = LM.GUI.PopupMenu(128, false)
	self.groupPopup:SetMenu(self.groupMenu)
	layout:AddChild(self.groupPopup)

	layout:AddChild(LM.GUI.StaticText(FA_SharedUtils.txX))
	self.textX = LM.GUI.TextControl(0, "00.0000", 0, LM.GUI.FIELD_FLOAT)
	self.textX:SetWheelInc(0.1)
	self.textX:SetValue(0)
	layout:AddChild(self.textX)

	layout:AddChild(LM.GUI.StaticText(FA_SharedUtils.txY))
	self.textY = LM.GUI.TextControl(0, "00.0000", 0, LM.GUI.FIELD_FLOAT)
	self.textY:SetWheelInc(0.1)
	self.textY:SetValue(0)
	layout:AddChild(self.textY)

	self.modifyBut= LM.GUI.Button(FA_SharedUtils.txTranslate, self.MODIFY)
	layout:AddChild(self.modifyBut)
	self.resetBut= LM.GUI.Button(FA_SharedUtils.txCopy0, self.RESET)
	layout:AddChild(self.resetBut)

	self.autoWeldCheck = LM.GUI.CheckBox(FA_SharedUtils.txAutoWeld, self.AUTOWELD)
	layout:AddChild(self.autoWeldCheck)

	self.autoFillCheck = LM.GUI.CheckBox(FA_SharedUtils.txAutoFill, self.AUTOFILL)
	layout:AddChild(self.autoFillCheck)

	self.lassoCheck = LM.GUI.CheckBox(FA_SharedUtils.txLasso, self.LASSO)
	layout:AddChild(self.lassoCheck)

	self.closestCheck = LM.GUI.CheckBox("Drag closest", self.CLOSEST)
	layout:AddChild(self.closestCheck)
end

function FA_TranslatePoints:UpdateWidgets(moho)
	self.autoWeldCheck:SetValue(FA_SharedUtils.autoWeld)
	self.autoFillCheck:SetValue(FA_SharedUtils.autoFill)
	self.lassoCheck:SetValue(FA_SharedUtils.lassoMode)
	self.closestCheck:SetValue(self.dragClosest)
	self.resetBut:Enable(moho.frame ~= 0)

	local mesh = moho:Mesh()

	if (mesh == nil) then
		self.groupMenu:RemoveAllItems()
		self.groupPopup:Enable(false)

		self.modifyBut:Enable(moho:CountSelectedBones() > 0)
		self.autoWeldCheck:Enable(false)
		self.autoFillCheck:Enable(false)
		self.closestCheck:Enable(false)
	else
		MOHO.BuildGroupMenu(self.groupMenu, mesh, self.SELECTITEM, self.DUMMY)
		self.groupPopup:Enable(true)

		self.modifyBut:Enable(moho:CountSelectedPoints() > 0)
		self.autoWeldCheck:Enable(true)
		self.autoFillCheck:Enable(true)
		self.closestCheck:Enable(true)
	end
end

function FA_TranslatePoints:HandleMessage(moho, view, msg)

	if (msg == self.MODIFY) then
		moho.document:PrepUndo(moho.layer)
		moho.document:SetDirty()

		local offset = LM.Vector2:new_local()
		offset:Set(self.textX:FloatValue(), self.textY:FloatValue())

		if (moho.layer:LayerType() == MOHO.LT_BONE) then
--bones
			FA_SharedUtils:PrepMoveBones(moho)
			FA_SharedUtils:TranslateBones(moho,offset)
			moho.layer:UpdateCurFrame()
			moho:NewKeyframe(CHANNEL_BONE)
		else
--points
			local mesh = moho:Mesh()
			if (mesh == nil) then return end

			mesh:PrepMovePoints()
			mesh:TranslatePoints(offset)
			moho:AddPointKeyframe(moho.frame)
			moho:NewKeyframe(CHANNEL_POINT)
		end

	elseif (msg == self.RESET) then

		if (moho.layer:LayerType() == MOHO.LT_BONE) then
--bones
			local skel = moho:Skeleton()
			if (skel == nil) then return end

			if (moho.frame > 0) then
				moho.document:PrepUndo(moho.layer)
				moho.document:SetDirty()
				self.addAdjust= true

				for i = 0, skel:CountBones() - 1 do
					local bone = skel:Bone(i)

					if (bone.fSelected) then
						bone.fAnimPos:SetValue(moho.frame, bone.fAnimPos:GetValue(0))
					elseif (bone.fParent >= 0) then

						if (skel:Bone(bone.fParent).fSelected) then
							bone.fAnimPos:SetValue(moho.frame, bone.fAnimPos:GetValue(0))
						end
					end
				end
				moho.layer:UpdateCurFrame()
				moho:NewKeyframe(CHANNEL_BONE)
			end
		else
--points
			if (moho.frame > 0) then
				local mesh = moho:Mesh()
				if (mesh == nil) then return end

				moho.document:PrepUndo(moho.layer)
				moho.document:SetDirty()

				for i = 0, mesh:CountPoints() - 1 do
					local pt = mesh:Point(i)
					if (pt.fSelected) then
						pt:SetPos(pt.fAnimPos:GetValue(0), moho.frame)
					end
				end
				moho:NewKeyframe(CHANNEL_POINT)
			end
		end
		self:UpdateWidgets(moho)
		self.textX:SetValue(0)
		self.textY:SetValue(0)
	elseif (msg == self.AUTOWELD) then
		FA_SharedUtils.autoWeld = self.autoWeldCheck:Value()
	elseif (msg == self.AUTOFILL) then
		FA_SharedUtils.autoFill = self.autoFillCheck:Value()
	elseif (msg == self.LASSO) then
		FA_SharedUtils.lassoMode = self.lassoCheck:Value()
	elseif (msg == self.CLOSEST) then
		self.dragClosest = self.closestCheck:Value()
	elseif (msg >= self.SELECTITEM) then
		local mesh = moho:Mesh()
		if (mesh == nil) then return end

		self.textX:SetValue(0)
		self.textY:SetValue(0)
		FA_SharedUtils:SelectItem(moho, mesh, msg - self.SELECTITEM)
	end
end
