-- **************************************************
-- Commonly used functions of the FA_ tools
-- **************************************************

-- **************************************************
-- General information about this script
-- **************************************************

-- Creator: Lost Marble, Modified by Fazek

FA_SharedUtils= {}

function FA_SharedUtils:Load_Prefs(prefs)

	if (not self.prefsLoad) then return end

	self.lassoMode = prefs:GetBool("FA_SharedUtils.lassoMode", false)
	self.autoClose = prefs:GetBool("FA_SharedUtils.autoClose", false)
	self.autoOutline = prefs:GetBool("FA_SharedUtils.autoOutline", true)
	self.autoFill = prefs:GetBool("FA_SharedUtils.autoFill", false)
	self.autoWeld = prefs:GetBool("FA_SharedUtils.autoWeld", false)
	self.sharpCorners = prefs:GetBool("FA_SharedUtils.sharpCorners", false)
	self.floodMaxStep = prefs:GetInt("FA_SharedUtils.floodMaxStep", 200)
	self.centerGravity = prefs:GetBool("FA_SharedUtils.centerGravity", false)
	self.prefsLoad= false
end

function FA_SharedUtils:Save_Prefs(prefs)
	if (self.prefsLoad) then return end

	prefs:SetBool("FA_SharedUtils.lassoMode", self.lassoMode)
	prefs:SetBool("FA_SharedUtils.autoClose", self.autoClose)
	prefs:SetBool("FA_SharedUtils.autoOutline", self.autoOutline)
	prefs:SetBool("FA_SharedUtils.autoFill", self.autoFill)
	prefs:SetBool("FA_SharedUtils.autoWeld", self.autoWeld)
	prefs:SetBool("FA_SharedUtils.sharpCorners", self.sharpCorners)
	prefs:SetInt("FA_SharedUtils.floodMaxStep", self.floodMaxStep)
	prefs:SetBool("FA_SharedUtils.centerGravity", self.centerGravity)
	self.prefsLoad= true
end

-- **************************************************
-- Recurring values
-- **************************************************

FA_SharedUtils.selRect= LM.Rect:new_local()
FA_SharedUtils.lastX= 0
FA_SharedUtils.lastY= 0
FA_SharedUtils.mouseMode= 0
FA_SharedUtils.selectionMode= 0 -- 0: general, 1: point mode, 2: shape mode, 3: bone mode
FA_SharedUtils.prefsLoad= true
FA_SharedUtils.drawMeMode= 0 -- 0: off, 1: preview shape
FA_SharedUtils.floodOverflow= false

-- prefs

FA_SharedUtils.lassoMode= true
FA_SharedUtils.autoClose = false
FA_SharedUtils.autoOutline = true
FA_SharedUtils.autoFill = false
FA_SharedUtils.autoWeld = false
FA_SharedUtils.sharpCorners = false
FA_SharedUtils.floodMaxStep = 200
FA_SharedUtils.centerGravity = false

-- text constants:

FA_SharedUtils.txSelectGroup = "Select Group"
FA_SharedUtils.txSelectBone = "Select Bone"
FA_SharedUtils.txX = "X:"
FA_SharedUtils.txY = "Y:"
FA_SharedUtils.txZ = "Z:"
FA_SharedUtils.txAngle = "Angle:"
FA_SharedUtils.txTranslate = "Translate"
FA_SharedUtils.txScale = "Scale"
FA_SharedUtils.txRotate = "Rotate"
FA_SharedUtils.txMinWidth = "Min width"
FA_SharedUtils.txMaxWidth = "Max width"
FA_SharedUtils.txModify = "Modify"
FA_SharedUtils.txReset = "Reset"
FA_SharedUtils.txCopy0 = "Copy Frm 0"
FA_SharedUtils.txDelete = "Delete"
FA_SharedUtils.txBindPoints = "Bind Points"
FA_SharedUtils.txCreateShape = "Create Shape"
FA_SharedUtils.txSetGroup = "Set Group"
FA_SharedUtils.txDeleteGroup = "Delete Group"
FA_SharedUtils.txAutoClose = "Auto close"
FA_SharedUtils.txAutoWeld = "Auto-weld"
FA_SharedUtils.txAutoFill = "Auto fill"
FA_SharedUtils.txAutoOutline = "Auto outline"
FA_SharedUtils.txSharpCorners = "Sharp corners"
FA_SharedUtils.txLasso = "Lasso select"
FA_SharedUtils.txCenterGravity = "Gravity centered"
FA_SharedUtils.txNone = "None"

-- **************************************************
-- The guts of this script
-- **************************************************

function FA_SharedUtils:MouseDown(moho,mouseEvent)

	if (self.selectionMode == 1) then
		local mesh= moho:Mesh()

		if (mesh == nil) then return end

		local ptID= mouseEvent.view:PickPoint(mouseEvent.pt)

		if (ptID >= 0) then
			moho.document:PrepUndo(moho.layer)
			mesh:SelectNone()
			mesh:Point(ptID).fSelected= true
			mouseEvent.view:DrawMe()
			self.mouseMode= -1
			return
		end

	elseif (self.selectionMode == 3) then
		local skel= moho:Skeleton()

		if (skel == nil) then return end

		local bnID= mouseEvent.view:PickBone(mouseEvent.pt,mouseEvent.vec,moho.layer,true)

		if (bnID >= 0) then
			moho.document:PrepUndo(moho.layer)

			for i= 0,skel:CountBones() - 1 do
				skel:Bone(i).fSelected= false
			end
			skel:Bone(bnID).fSelected= true
			moho:UpdateBonePointSelection()
			mouseEvent.view:DrawMe()
			self.mouseMode= -1
			return
		end
	end

	local lassoMode= self.lassoMode

	if (mouseEvent.ctrlKey) then
		lassoMode= not lassoMode
	end

	if (lassoMode) then
--set lasso mode
		self.lassoList = { { mouseEvent.startPt.x,mouseEvent.startPt.y } }
		self.lastX = mouseEvent.startPt.x
		self.lastY = mouseEvent.startPt.y
		self.mouseMode= 1
	else
--set rectangle mode
		self.selRect.left = mouseEvent.startPt.x
		self.selRect.top = mouseEvent.startPt.y
		self.selRect.right = mouseEvent.pt.x
		self.selRect.bottom = mouseEvent.pt.y
		mouseEvent.view:Graphics():SelectionRect(self.selRect)
		self.mouseMode= 2
	end
	mouseEvent.view:DrawMe()
end

function FA_SharedUtils:MouseMoved(moho,mouseEvent)

	if (self.mouseMode == 0) then return end

--mouse drag or just a click?

	if (self.mouseMode < 3) then

		if (self.mouseMode == -1) then
--forced fallback
			self.mouseMode= 0
			self.lassoList= nil
			return
		end
		local x= mouseEvent.startPt.x - mouseEvent.pt.x
		local y= mouseEvent.startPt.y - mouseEvent.pt.y

		if ((x*x + y*y) < 9) then return end

		if (self.selectionMode == 1 or self.selectionMode == 3) then
--pass back to the user
			self.mouseMode= 0
			self.lassoList= nil
			return
		end
		self.mouseMode= self.mouseMode + 2
	end
	local gfx= mouseEvent.view:Graphics()

	if (self.mouseMode == 3) then
		gfx:SetSmoothing(true)
		gfx:Push()
		local gm= gfx:CurrentTransform()
		gm:Invert()
		gfx:ApplyMatrix(gm)
		gfx:SetColor(255,0,0)
		gfx:MoveTo(self.lastX,self.lastY)
		gfx:LineTo(mouseEvent.pt.x, mouseEvent.pt.y)
		gfx:Pop()
		gfx:SetSmoothing(false)
		table.insert(self.lassoList, { mouseEvent.pt.x,mouseEvent.pt.y })
		self.lastX= mouseEvent.pt.x
		self.lastY= mouseEvent.pt.y
	else
		gfx:SelectionRect(self.selRect)
		self.selRect.right= mouseEvent.pt.x
		self.selRect.bottom= mouseEvent.pt.y
		gfx:SelectionRect(self.selRect)
	end
	mouseEvent.view:Refresh()
end

function FA_SharedUtils:MouseUp(moho,mouseEvent)

	if (self.mouseMode < 1) then
		self.lassoList= nil
		return
	end
	local skel,mesh
	local t= false

	if (self.mouseMode < 3) then

		local bnID= -1

		if (moho.layer:LayerType() == MOHO.LT_BONE) then
			skel= moho:Skeleton()

			if (skel == nil) then return end

			bnID= mouseEvent.view:PickBone(mouseEvent.pt,mouseEvent.vec,moho.layer,true)

			if (bnID < 0) then

				if (moho:CountSelectedBones() > 0) then
					moho.document:PrepUndo(moho.layer)

					for i= 0,skel:CountBones() - 1 do
						skel:Bone(i).fSelected= false
					end
				end
			end
		else
			mesh= moho:Mesh()

			if (mesh == nil) then return end

			local ptID= mouseEvent.view:PickPoint(mouseEvent.pt)
			local cveID,segID= -1,-1
			local pt,ptMax
--clicked on a point?

			if (ptID >= 0) then
				moho.document:PrepUndo(moho.layer)
				pt= mesh:Point(ptID)
				pt.fSelected= not pt.fSelected
			else

--clicked on an edge?
				cveID, segID= mouseEvent.view:PickEdge(mouseEvent.pt, cveID,segID)

				if (cveID >= 0) then
					local cve= mesh:Curve(cveID)

					ptMax= cve:CountPoints() - 1

					if (mouseEvent.ctrlKey) then
	--whole curve
		--pass1: test
						for i=0, ptMax do
							pt= cve:Point(i)

							if (not pt.fSelected) then
								t= true
								break
							end
						end
		--pass2: set
						moho.document:PrepUndo(moho.layer)

						for i= 0, ptMax do
							pt= cve:Point(i)
							pt.fSelected= t
						end
					else
	--part of the curve
		--pass1: test (all points selected?)
						local ovf2,ovf1
						ovf1,t= self:TestSegments(cve,segID,0,-1,t)

						if (segID < ptMax) then
							ovf2,t= self:TestSegments(cve,segID + 1,ptMax,1,t)
						else
							ovf2= -1
						end

						if (cve.fClosed) then
			--roundover
							if (ovf1 < 0 and ovf2 >= 0) then
								ovf2,t= self:TestSegments(cve,ptMax,ovf2,-1,t)
							elseif (ovf2 < 0) then
								ovf1,t= self:TestSegments(cve,0,ovf1,1,t)
							end
						end
		--pass2: set
						moho.document:PrepUndo(moho.layer)

						ovf1= self:SetSegments(cve,segID,0,-1,segID + 1,t)

						if (segID < ptMax) then
							ovf2= self:SetSegments(cve,segID + 1,ptMax,1,segID,t)
						else
							ovf2= -1
						end

						if (cve.fClosed) then
			--roundover
							if (ovf1 < 0 and ovf2 >= 0) then
								self:SetSegments(cve,ptMax,ovf2,-1,nil,t)
							elseif (ovf2 < 0) then
								self:SetSegments(cve,0,ovf1,1,nil,t)
							end
						end
					end
				else
					local shpID= -1
--clicked on a parent bone?
					skel= moho:ParentSkeleton()

					if (skel ~= nil) then
						bnID= mouseEvent.view:PickBone(mouseEvent.pt,mouseEvent.vec,moho.layer:Parent(),true)
				--currently the following has no effect:
				--if (bnID >= 0) then moho.document:PrepUndo(moho.layer:Parent()) end

						if (bnID < 0 and self.selectionMode < 2) then
							shpID= mouseEvent.view:PickShape(mouseEvent.pt)
						end

					elseif (self.selectionMode < 2) then
						shpID= mouseEvent.view:PickShape(mouseEvent.pt)
					end
--clicked on a shape?
					if (shpID >= 0) then
						local shp= mesh:Shape(shpID)
						local lastID
						local cve
						local edgMax= shp:CountEdges() - 1

						ptMax= 0
	--pass1: test
						for i= 0,edgMax do
							cveID, segID= shp:GetEdge(i,cveID,segID)

							if (cveID ~= lastID) then
								lastID= cveID
								cve= mesh:Curve(cveID)
								ptMax= cve:CountPoints() - 1
							end

							if (not cve:Point(segID).fSelected) then
								t= true
								break
							end

							if (segID < ptMax) then
								if (not cve:Point(segID+1).fSelected) then
									t= true
									break
								end
							elseif (not cve:Point(0).fSelected) then
								t= true
								break
							end
						end
	--pass2: set
						moho.document:PrepUndo(moho.layer)

						for i= 0, edgMax do
							cveID, segID= shp:GetEdge(i,cveID,segID)

							if (cveID ~= lastID) then
								lastID= cveID
								cve= mesh:Curve(cveID)
								ptMax= cve:CountPoints() - 1
							end
							cve:Point(segID).fSelected= t

							if (segID < ptMax) then
								cve:Point(segID+1).fSelected= t
							else
								cve:Point(0).fSelected= t
							end
						end
						shp.fSelected= not shp.fSelected
					else
						if (self.selectionMode == 2) then
							t= self:FloodSelection(moho,mouseEvent)
						end

						if (not t) then

							if (bnID < 0) then
								moho.document:PrepUndo(moho.layer)
								mesh:SelectNone()
							end

							if (skel ~= nil) then

								for i= 0,skel:CountBones() - 1 do
									skel:Bone(i).fSelected= false
								end
							end
						end
					end
				end
			end
		end

		if (bnID >= 0) then
			moho.document:PrepUndo(moho.layer)
			local bone= skel:Bone(bnID)
			bone.fSelected= not bone.fSelected
			moho:UpdateBonePointSelection()
		end
		mouseEvent.view:DrawMe()
		self.lassoList= nil
		self.mouseMode= 0
		moho:UpdateSelectedChannels()
		return
	end

	local gfx= mouseEvent.view:Graphics()
	local vec= LM.Vector2:new_local()
	local screenPt= LM.Point:new_local()

	local tMat= LM.Matrix:new_local()
	moho.layer:GetFullTransform(moho.frame,tMat,moho.document)

	if (self.mouseMode == 3) then
--lasso
		gfx:SetSmoothing(true)
		gfx:Push()
		local gm= gfx:CurrentTransform()
		gm:Invert()
		gfx:ApplyMatrix(gm)
		gfx:SetColor(255,0,0)
		gfx:MoveTo(self.lastX,self.lastY)
		gfx:LineTo(mouseEvent.startPt.x,mouseEvent.startPt.y)
		gfx:Pop()
		gfx:SetSmoothing(false)
		mouseEvent.view:Refresh()
		LM.Snooze(100)

		gfx:Clear(0, 0, 0, 0)
		gfx:Push()
		gfx:ApplyMatrix(gm)
		gfx:SetColor(255,255,255,255)
		gfx:BeginShape()

		local tabN= table.getn(self.lassoList)
		local vec2= LM.Vector2:new_local()

		for i= 1, table.getn(self.lassoList) - 1 do
			vec:Set(self.lassoList[i][1], self.lassoList[i][2])
			vec2:Set(self.lassoList[i + 1][1], self.lassoList[i + 1][2])
			gfx:AddLine(vec,vec2)
		end
		vec:Set(self.lassoList[1][1],self.lassoList[1][2])
		gfx:AddLine(vec2,vec)
		gfx:EndShape()
		gfx:Pop()

		gfx:Push()
		gfx:ApplyMatrix(tMat)

		if (moho.layer:LayerType() == MOHO.LT_BONE) then
--bones
			skel= moho:Skeleton()

			if (skel == nil) then return end
	--pass1: test
			for i = 0, skel:CountBones() - 1 do
				local bone= skel:Bone(i)

				if (not bone.fSelected) then
					vec:Set(0,0)

					if (moho.frame == 0) then
						tMat:Set(bone.fRestMatrix)
					else
						tMat:Set(bone.fMovedMatrix)
					end
					tMat:Transform(vec)
					gfx:WorldToScreen(vec,screenPt)

					if (gfx:IsFullWhite(screenPt)) then
						vec2:Set(bone.fLength,0)
						tMat:Transform(vec2)
						gfx:WorldToScreen(vec2,screenPt)

						if (gfx:IsFullWhite(screenPt)) then
							vec2= (vec + vec2) / 2
							gfx:WorldToScreen(vec2,screenPt)

							if (gfx:IsFullWhite(screenPt)) then
								t= true
								break
							end
						end
					end
				end
			end
	--pass2: set
			moho.document:PrepUndo(moho.layer)

			for i = 0, skel:CountBones() - 1 do
				local bone= skel:Bone(i)

				if (bone.fSelected ~= t) then
					vec:Set(0,0)

					if (moho.frame == 0) then
						tMat:Set(bone.fRestMatrix)
					else
						tMat:Set(bone.fMovedMatrix)
					end
					tMat:Transform(vec)
					gfx:WorldToScreen(vec,screenPt)

					if (gfx:IsFullWhite(screenPt)) then
						vec2:Set(bone.fLength,0)
						tMat:Transform(vec2)
						gfx:WorldToScreen(vec2,screenPt)

						if (gfx:IsFullWhite(screenPt)) then
							vec2= (vec + vec2) / 2
							gfx:WorldToScreen(vec2,screenPt)

							if (gfx:IsFullWhite(screenPt)) then
								bone.fSelected= t
							end
						end
					end
				end
			end
		else
--points
			mesh= moho:Mesh()

			if (mesh == nil) then return end
	--pass1: test
			for i = 0, mesh:CountPoints() - 1 do
				local pt= mesh:Point(i)

				if (not pt.fSelected) then
					gfx:WorldToScreen(pt.fPos,screenPt)

					if (gfx:IsFullWhite(screenPt)) then
						t= true
						break
					end
				end
			end
	--pass2: set
			moho.document:PrepUndo(moho.layer)

			for i = 0, mesh:CountPoints() - 1 do
				local pt= mesh:Point(i)

				if (pt.fSelected ~= t) then
					gfx:WorldToScreen(pt.fPos,screenPt)

					if (gfx:IsFullWhite(screenPt)) then
						pt.fSelected= t
					end
				end
			end
		end
	else
--rectangle
		self.selRect:Normalize()

		gfx:Push()
		gfx:ApplyMatrix(tMat)

		if (moho.layer:LayerType() == MOHO.LT_BONE) then
--bones
			skel= moho:Skeleton()

			if (skel == nil) then return end
	--pass1: test
			for i= 0, skel:CountBones() - 1 do
				local bone= skel:Bone(i)

				if (not bone.fSelected) then
					vec:Set(0,0)

					if (moho.frame == 0) then
						tMat:Set(bone.fRestMatrix)
					else
						tMat:Set(bone.fMovedMatrix)
					end
					tMat:Transform(vec)
					gfx:WorldToScreen(vec,screenPt)

					if (self.selRect:Contains(screenPt)) then
						vec:Set(bone.fLength,0)
						tMat:Transform(vec)
						gfx:WorldToScreen(vec,screenPt)

						if (self.selRect:Contains(screenPt)) then
							t= true
							break
						end
					end
				end
			end
	--pass2: set
			moho.document:PrepUndo(moho.layer)

			for i= 0, skel:CountBones() - 1 do
				local bone= skel:Bone(i)

				if (bone.fSelected ~= t) then
					vec:Set(0,0)

					if (moho.frame == 0) then
						tMat:Set(bone.fRestMatrix)
					else
						tMat:Set(bone.fMovedMatrix)
					end
					tMat:Transform(vec)
					gfx:WorldToScreen(vec,screenPt)

					if (self.selRect:Contains(screenPt)) then
						vec:Set(bone.fLength,0)
						tMat:Transform(vec)
						gfx:WorldToScreen(vec,screenPt)

						if (self.selRect:Contains(screenPt)) then
							bone.fSelected= t
						end
					end
				end
			end
		else
--points
			mesh= moho:Mesh()

			if (mesh == nil) then return end
	--pass1: test
			for i= 0, mesh:CountPoints() - 1 do
				local pt= mesh:Point(i)

				if (not pt.fSelected) then
					gfx:WorldToScreen(pt.fPos,screenPt)

					if (self.selRect:Contains(screenPt)) then
						t= true
						break
					end
				end
			end
	--pass2: set
			moho.document:PrepUndo(moho.layer)

			for i= 0, mesh:CountPoints() - 1 do
				local pt= mesh:Point(i)

				if (pt.fSelected ~= t) then
					gfx:WorldToScreen(pt.fPos,screenPt)

					if (self.selRect:Contains(screenPt)) then
						pt.fSelected = t
					end
				end
			end
		end
	end
	gfx:Pop()

	mouseEvent.view:DrawMe()
	self.lassoList= nil
	self.mouseMode= 0
	moho:UpdateSelectedChannels()
end

function FA_SharedUtils:KeyDown(moho,keyEvent)
	local mesh= moho:Mesh()
	if (mesh == nil) then return end

	if ((keyEvent.keyCode == LM.GUI.KEY_DELETE) or (keyEvent.keyCode == LM.GUI.KEY_BACKSPACE)) then

		if (moho:CountSelectedPoints() > 0) then
			moho.document:PrepUndo(moho.layer)
			moho.document:SetDirty()
			MOHO.DeleteSelectedPoints(mesh)
			keyEvent.view:DrawMe()
		end
	end
end

function FA_SharedUtils:Draw_Me(moho, view, mode)

	if (mode < 0) then
		mode= self.drawMeMode
	end

	if (mode == 1) then
		local gfx = view:Graphics()
		local matrix = LM.Matrix:new_local()

		moho.layer:GetFullTransform(moho.frame, matrix, moho.document)
		gfx:Push()
		gfx:ApplyMatrix(matrix)

		view:DrawPreviewShape()
		gfx:Pop()
	end
end

-- local functions

function FA_SharedUtils:ClearUnusedPoint(pt)
	local segID,cve= -1
	local f= true

	for i= 0, pt:CountCurves() - 1 do
		cve,segID= pt:Curve(i,segID)
		local segMax= cve:CountSegments() - 1

		if (segID <= segMax and cve:IsSegmentSelected(segID)) then
			f= false
			break
		end

		if (segID > 0) then

			if (cve:IsSegmentSelected(segID - 1)) then
				f= false
				break
			end
		else
	--handle roundovers

			if (cve.fClosed) then
	--closed curve
				if (cve:IsSegmentSelected(segMax)) then
					f= false
					break
				end
			end
		end
	end

	if (f) then
		pt.fSelected= false
	end
end

function FA_SharedUtils:SetSegments(cve,min,max,step,e,t)
	local pt

--on a part of a curve, set points to selection state 't'

	for i= min,max,step do
		pt= cve:Point(i)

		if (pt:CountCurves() > 1) then

			if (t or not pt.fSelected) then
				pt.fSelected= t
				return i
			end

			if (i == min and e) then
				local npt= cve:Point(e)
				local f= npt.fSelected
				npt.fSelected= false
				self:ClearUnusedPoint(pt)
				npt.fSelected= f
			else
				self:ClearUnusedPoint(pt)
			end
			return i
		end
		pt.fSelected= t
	end
	return -1
end

function FA_SharedUtils:TestSegments(cve,min,max,step,t)
	local pt

	if (t) then
		return -1,true
	end

--on a part of a curve, return true if unselected points exists

	for i= min,max,step do
		pt= cve:Point(i)

		if (not pt.fSelected) then
			return -1,true
		end

		if (pt:CountCurves() > 1) then
			return i,false
		end
	end
	return -1,false
end

function FA_SharedUtils:CollectSegments(mesh)

--collect all curve segments between wended or end points, calculating curve length squares

	local segList= {}

	if (not mesh) then
		return segList
	end

	local pt,cve,min,minID,cveN,ol,lastPt,x,y

	for i= 0, mesh:CountPoints() - 1 do
	--a trick to retrieve mesh ID
		pt= mesh:Point(i)
		pt.fTempPos.x= i
	end

	for i= 0, mesh:CountCurves() - 1 do
		cve= mesh:Curve(i)
		cveN= cve:CountPoints() - 1
		pt= cve:Point(0)
		minID= pt.fTempPos.x

		if (cve.fClosed) then
--roundover?
			min= -1
			ol= 0
			lastPt= cve:Point(0)

			for j= cveN, 0, -1 do
				pt= cve:Point(j)
				x= pt.fPos.x - lastPt.fPos.x
				y= pt.fPos.y - lastPt.fPos.y
				ol= ol + (x * x) + (y * y)

				if (pt:CountCurves() > 1) then
					minID= pt.fTempPos.x
					cveN= j
					lastPt= cve:Point(0)

					for k= 0, j do
						pt= cve:Point(k)
						x= pt.fPos.x - lastPt.fPos.x
						y= pt.fPos.y - lastPt.fPos.y
						ol= ol + (x * x) + (y * y)

						if (pt:CountCurves() > 1) then
							table.insert(segList,{ i,j,k,minID,pt.fTempPos.x,true,true,ol })
							min= k
							minID= pt.fTempPos.x
							break
						end
					end
					break
				end
			end

			if (min < 0) then
				table.insert(segList,{ i,0,0,minID,minID,true,true,ol })
				min= cveN
			end
		else
			min= 0
		end

		if (min < cveN) then
			ol= 0
			lastPt= cve:Point(min)

			for j= min + 1, cveN do
				pt= cve:Point(j)
				x= pt.fPos.x - lastPt.fPos.x
				y= pt.fPos.y - lastPt.fPos.y
				ol= ol + (x * x) + (y * y)

				if (pt:CountCurves() > 1) then
					table.insert(segList,{ i,min,j,minID,pt.fTempPos.x,true,true,ol })
					min= j
					minID= pt.fTempPos.x
					ol= 0
					lastPt= pt
				end
			end

			if (min < cveN) then
				table.insert(segList,{ i,min,cveN,minID,pt.fTempPos.x,true,true,ol })
			end
		end
	end
	return segList
end

function FA_SharedUtils:FloodSelection(moho,mouseEvent)
	local maxStep= self.floodMaxStep

	if (maxStep < 1) then
		return false
	end
	local mesh= moho:Mesh()
	local segList= self:CollectSegments(mesh)
	local tabN= table.getn(segList)

	if (tabN < 1) then
		return false
	end

--find all closed curve variations

	local chain= { { 0,0 } }
	local best= nil
	local chN= 1
	local a1,a2,b1,b2
	local i= tabN + 1
	local bestN= i
	local bestOl
	local first= 0
	local t= false
	local gfx= mouseEvent.view:Graphics()
	local cve,min,max,cveID,pt1,pt2,hVec1,hVec2,hVec3
	local screenPt= LM.Point:new_local()

	while true do

		while true do

			if (i > tabN) then
	--pop
				if (chN == 1) then
					first= first + 1

					if (first > tabN) then

						if (not best) then return false end
	--finished: use the best match to select
						t= false
		--pass1: test
						for j= 1, bestN do
							cveID= best[j]
							cve= mesh:Curve(segList[cveID][1])
							min= segList[cveID][2]
							max= segList[cveID][3]

							if (max <= min) then
			--roundover
								for k= min, cve:CountSegments() - 1 do
									if (not cve:Point(k).fSelected) then
										t= true
										break
									end
								end

								if (not t) then

									for k= 0, max do
										if (not cve:Point(k).fSelected) then
											t= true
											break
										end
									end
								end
							else
								for k= min, max do

									if (not cve:Point(k).fSelected) then
										t= true
										break
									end
								end
							end
						end
		--pass2: set
						moho.document:PrepUndo(moho.layer)

						for j= 1, bestN do
							cveID= best[j]
							cve= mesh:Curve(segList[cveID][1])
							min= segList[cveID][2]
							max= segList[cveID][3]

							if (max <= min) then
			--roundover
								for k= min, cve:CountSegments() - 1 do
									cve:Point(k).fSelected= t
								end

								for k= 0, max do
									cve:Point(k).fSelected= t
								end
							else
								for k= min, max do
									cve:Point(k).fSelected= t
								end
							end
						end
						return true
					end
					a1= segList[first][4]
					a2= segList[first][5]
					segList[first][6]= false
					chain[1][1]= first
					chain[1][2]= -1
					i= first

					if (a1 == a2) then
	--closed segment
						t= true
						break
					end
				else
					i= chain[chN][1]
					a2= chain[chN][2]
					segList[i][6]= true
					table.remove(chain)
					chN= chN - 1
				end

			elseif (segList[i][6] and segList[i][7]) then
				b1= segList[i][4]
				b2= segList[i][5]

				if (b1 == a2) then
	--push
					table.insert(chain,{ i, a2 })
			 		segList[i][6]= false
					chN= chN + 1

					if (b2 == a1) then
						t= true
						break
					end
					a2= b2
					i= first
				elseif (b2 == a2) then
	--push
					table.insert(chain,{ i, a2 })
					segList[i][6]= false
					chN= chN + 1

					if (b1 == a1) then
						t= true
						break
					end
					a2= b1
					i= first
				end
			end
			i= i + 1
		end

--fill the shape for testing

		if (t) then

			if (maxStep < 1) then
				self.floodOverflow= true
				moho:Click()
				return false
			end
			maxStep= maxStep - 1

			gfx:Clear(0, 0, 0, 0)
			gfx:SetColor(255,255,255,255)
			gfx:BeginShape()

			for j= 1, chN do
				cveID= chain[j][1]
				cve= mesh:Curve(segList[cveID][1])
				min= segList[cveID][2]
				max= segList[cveID][3]

				if (max <= min) then
		--roundover
					pt1= cve:Point(min)
					hVec1= cve:PointOnSegment(min,0.25)
					hVec2= cve:PointOnSegment(min,0.5)
					hVec3= cve:PointOnSegment(min,0.75)

					for k= min + 1, cve:CountSegments() - 1 do
						pt2= cve:Point(k)
						gfx:AddLine(pt1.fPos,hVec1)
						gfx:AddLine(hVec1,hVec2)
						gfx:AddLine(hVec2,hVec3)
						gfx:AddLine(hVec3,pt2.fPos)
						hVec1= cve:PointOnSegment(k,0.25)
						hVec2= cve:PointOnSegment(k,0.5)
						hVec3= cve:PointOnSegment(k,0.75)
						pt1= pt2
					end

					for k= 0, max - 1 do
						pt2= cve:Point(k)
						gfx:AddLine(pt1.fPos,hVec1)
						gfx:AddLine(hVec1,hVec2)
						gfx:AddLine(hVec2,hVec3)
						gfx:AddLine(hVec3,pt2.fPos)
						hVec1= cve:PointOnSegment(k,0.25)
						hVec2= cve:PointOnSegment(k,0.5)
						hVec3= cve:PointOnSegment(k,0.75)
						pt1= pt2
					end
					pt2= cve:Point(max)
					gfx:AddLine(pt1.fPos,hVec1)
					gfx:AddLine(hVec1,hVec2)
					gfx:AddLine(hVec2,hVec3)
					gfx:AddLine(hVec3,pt2.fPos)
				else
					pt1= cve:Point(min)
					hVec1= cve:PointOnSegment(min,0.25)
					hVec2= cve:PointOnSegment(min,0.5)
					hVec3= cve:PointOnSegment(min,0.75)

					for k= min + 1, max - 1 do
						pt2= cve:Point(k)
						gfx:AddLine(pt1.fPos,hVec1)
						gfx:AddLine(hVec1,hVec2)
						gfx:AddLine(hVec2,hVec3)
						gfx:AddLine(hVec3,pt2.fPos)
						hVec1= cve:PointOnSegment(k,0.25)
						hVec2= cve:PointOnSegment(k,0.5)
						hVec3= cve:PointOnSegment(k,0.75)
						pt1= pt2
					end
					pt2= cve:Point(max)
					gfx:AddLine(pt1.fPos,hVec1)
					gfx:AddLine(hVec1,hVec2)
					gfx:AddLine(hVec2,hVec3)
					gfx:AddLine(hVec3,pt2.fPos)
				end
			end
			gfx:EndShape()

			if (gfx:IsFullWhite(mouseEvent.pt)) then
--clicked inside: remove outside edges
				for j= first + 1, tabN do

					if (segList[j][6]) then
						cve= mesh:Curve(segList[j][1])
						min= segList[j][2]
						max= segList[j][3]
						segList[j][7]= false

						if (max <= min) then
			--roundover
							if (max == 0 and (min >= cve:CountSegments() - 1)) then
								hVec1= cve:PointOnSegment(max,0.5)
								gfx:WorldToScreen(hVec1,screenPt)

								if (gfx:IsFullWhite(screenPt)) then
									segList[j][7]= true
								end
							else
								t= true

								for k= min + 1,cve:CountSegments() - 1 do
									gfx:WorldToScreen(cve:Point(k).fPos,screenPt)

									if (gfx:IsFullWhite(screenPt)) then
										segList[j][7]= true
										t= false
										break
									end
								end

								if (t) then
									for k= 0, max - 1 do
										gfx:WorldToScreen(cve:Point(k).fPos,screenPt)

										if (gfx:IsFullWhite(screenPt)) then
											segList[j][7]= true
											break
										end
									end
								end
							end
						else
							if (min + 1 >= max) then
								hVec1= cve:PointOnSegment(min,0.5)
								gfx:WorldToScreen(hVec1,screenPt)

								if (gfx:IsFullWhite(screenPt)) then
									segList[j][7]= true
								end
							else
								for k= min + 1, max - 1 do
									gfx:WorldToScreen(cve:Point(k).fPos,screenPt)

									if (gfx:IsFullWhite(screenPt)) then
										segList[j][7]= true
										break
									end
								end
							end
						end
					end
				end
				local ol= 0

				if (best) then
--only if the outline is smaller
					for j= 1, chN do
						i= chain[j][1]
						ol= ol + segList[i][8]
					end

					if (ol <= bestOl) then
						best= {}

						for j= 1, chN do
							i= chain[j][1]
							table.insert( best, i )
							segList[i][7]= true
						end
						bestOl= ol
						bestN= chN
					end
				else
					best= {}

					for j= 1, chN do
						i= chain[j][1]
						table.insert( best, i )
						segList[i][7]= true
						ol= ol + segList[i][8]
					end
					bestN= chN
					bestOl= ol
				end
			else
-- outside: remove internal edges
				for j= first + 1, tabN do

					if (segList[j][6] and segList[j][7]) then
						cve= mesh:Curve(segList[j][1])
						min= segList[j][2]
						max= segList[j][3]
						segList[j][7]= false

						if (max <= min) then
			--roundover
							if (max == 0 and (min >= cve:CountSegments() - 1)) then
								hVec1= cve:PointOnSegment(max,0.5)
								gfx:WorldToScreen(hVec1,screenPt)

								if (not gfx:IsFullWhite(screenPt)) then
									segList[j][7]= true
								end
							else
								t= true

								for k= min + 1,cve:CountSegments() - 1 do
									gfx:WorldToScreen(cve:Point(k).fPos,screenPt)

									if (not gfx:IsFullWhite(screenPt)) then
										segList[j][7]= true
										t= false
										break
									end
								end

								if (t) then
									for k= 0, max - 1 do
										gfx:WorldToScreen(cve:Point(k).fPos,screenPt)

										if (not gfx:IsFullWhite(screenPt)) then
											segList[j][7]= true
											break
										end
									end
								end
							end
						else
							if (min + 1 >= max) then
								hVec1= cve:PointOnSegment(min,0.5)
								gfx:WorldToScreen(hVec1,screenPt)

								if (not gfx:IsFullWhite(screenPt)) then
									segList[j][7]= true
								end
							else
								for k= min + 1, max - 1 do
									gfx:WorldToScreen(cve:Point(k).fPos,screenPt)

									if (not gfx:IsFullWhite(screenPt)) then
										segList[j][7]= true
										break
									end
								end
							end
						end
					end
				end

				for j= 1, chN do
					segList[chain[j][1]][7]= true
				end
			end
			t= false
			i= tabN + 1
		end
	end
--never executes this point
end

function FA_SharedUtils:CreateShape(moho, mesh, ptID)
	mesh:Point(ptID).fSelected = true
	mesh:SelectConnected()

-- filled shape
	self.drawMeMode = 1
	local shapeID = moho:CreateShape(true)
	self.drawMeMode = 0

	if (shapeID >= 0) then
		mesh:Shape(shapeID).fSelected = true
		moho:NewKeyframe(CHANNEL_FILL)
		moho:NewKeyframe(CHANNEL_LINE)
		moho:UpdateSelectedChannels()
	else
		mesh:SelectNone()
		mesh:Point(ptID).fSelected = true
	end
end

function FA_SharedUtils:WeldPoints(moho, view, dragging, selID)
	local mesh = moho:Mesh()
	if (mesh == nil) then
		return false
	end

-- try welding to the closest point

	if (not dragging) then
		selID = -1
		local numSel= 0

		for i = 0, mesh:CountPoints() - 1 do
			local pt = mesh:Point(i)

			if (pt.fSelected) then
				numSel = numSel + 1
				selID = i

				if (numSel > 1) then
					break
				end
			end
		end

		if (numSel > 1) then
			selID = -1
		end
		if (selID < 0) then
			return false
		end
		moho.document:PrepUndo(moho.layer)
	end

	if (selID < 0) then
		return false
	end
	moho.document:SetDirty()

	local testVec1 = LM.Vector2:new_local()
	local testVec2 = LM.Vector2:new_local()

	testVec1:Set(mesh:Point(selID).fPos)
	local closestID = mesh:ClosestPoint(testVec1, selID)
	testVec2:Set(mesh:Point(closestID).fPos)

-- convert the two chosen points to pixel coordinates
-- this is to make sure that they are actually close enough in screen space

	local p1 = LM.Point:new_local()
	local p2 = LM.Point:new_local()
	local m = LM.Matrix:new_local()
	moho.layer:GetFullTransform(moho.frame, m, moho.document)
	m:Transform(testVec1)
	m:Transform(testVec2)
	view:Graphics():WorldToScreen(testVec1, p1)
	view:Graphics():WorldToScreen(testVec2, p2)
	p1.x = p1.x - p2.x
	p1.y = p1.y - p2.y

	if ((p1.x*p1.x + p1.y*p1.y) >= 100) then
		return false
	end

-- found one close enough to weld

	if (mesh:ArePointsAdjacent(selID, closestID)) then
-- this avoids welding adjacent points, particularly the end point of a curve to the previous point

		return true
	end

	if (mesh:WeldPoints(selID, closestID, moho.frame)) then
		moho:Click()

		if (selID < closestID) then
			closestID = closestID - 1
		end

		if (self.autoFill) then
			self:CreateShape(moho, mesh, closestID)
		end
		return true
	end
	return false
end

function FA_SharedUtils:TryWeldPoints(moho, view, dragging, selID)

	if (not self:WeldPoints(moho, view, dragging, selID)) then

-- Failed to weld this endpoint to another point.
-- Instead, try welding it to the middle of a nearby curve.

		local mesh = moho:Mesh()
		if (mesh == nil) then return end

		local m = LM.Matrix:new_local()
		local v = LM.Vector2:new_local()
		local screenPt = LM.Point:new_local()
		local curveID = -1
		local segID = -1

		moho.layer:GetFullTransform(moho.frame, m, moho.document)
		v:Set(mesh:Point(selID).fPos)
		m:Transform(v)
		view:Graphics():WorldToScreen(v, screenPt)
		curveID, segID = mesh:Point(selID):GetEndpointEdge(curveID, segID)
		curveID, segID = view:PickEdge(screenPt, curveID, segID)

		if (curveID >= 0) then
-- add a point in the middle of some curve
			if (not mesh:Curve(curveID):IsPointOnSegment(selID, segID)) then
	-- don't weld the point back on itself
				mesh:AddPoint(mesh:Point(selID).fPos, curveID, segID, moho.frame)

				if (mesh:WeldPoints(selID, mesh:CountPoints() - 1, moho.frame)) then
					moho:Click()

					if (self.autoFill) then
						self:CreateShape(moho, mesh, mesh:CountPoints() - 1)
					end
				end
			end
		end
	end
end

function FA_SharedUtils:TranslateStep(moho,mesh,view,dx,dy)
	local id = moho:CountSelectedPoints()

	if (id < 1) then return end

	moho.document:PrepUndo(moho.layer)
	moho.document:SetDirty()
	mesh:PrepMovePoints()

	local screenPt = LM.Point:new_local()
	local m = LM.Matrix:new_local()
	screenPt:Set(0,0)
	moho.layer:GetFullTransform(moho.frame,m,moho.document)
	local vec1 = view:Point2Vec(screenPt,m)
	screenPt:Set(dx,dy)
	local vec2 = view:Point2Vec(screenPt,m)
	vec2= vec2 - vec1
	mesh:TranslatePoints(vec2)
	moho:AddPointKeyframe(moho.frame)
	view:DrawMe()

--no welding: too many side effects for the user

end

function FA_SharedUtils:SelectedCenter(moho,flag)
	local vecL,vecH= self:SelectedBounds(moho,flag)
	return (vecL + vecH) / 2
end

function FA_SharedUtils:SelectedBounds(moho,flag)
	local vecL = LM.Vector2:new_local()
	local vecH = LM.Vector2:new_local()
	local vecC = LM.Vector2:new_local()
	vecL:Set(0,0)
	vecH:Set(0,0)
	local cnt= nil

	local mesh = moho:Mesh()

	if (mesh ~= nil) then
		local vec = LM.Vector2:new_local()

		for i = 0, mesh:CountPoints() - 1 do
			local pt= mesh:Point(i)

			if (pt.fSelected) then
				vec:Set(pt.fPos)

				if (cnt) then
					if (vecL.x > vec.x) then
						vecL.x= vec.x
					elseif (vecH.x < vec.x) then
						vecH.x= vec.x
					end
					if (vecL.y > vec.y) then
						vecL.y= vec.y
					elseif (vecH.y < vec.y) then
						vecH.y= vec.y
					end
					vecC= vecC + vec
					cnt= cnt + 1
				else
					vecL:Set(vec)
					vecH:Set(vec)
					vecC:Set(vec)
					cnt= 1
				end
			end
		end
	end

	if (cnt and flag) then
--center correction
		vecC= ((vecC*2)/cnt) - vecL - vecH

		if (vecC.x < 0) then
			vecL.x= vecL.x + vecC.x
		else
			vecH.x= vecH.x + vecC.x
		end

		if (vecC.y < 0) then
			vecL.y= vecL.y + vecC.y
		else
			vecH.y= vecH.y + vecC.y
		end
	end
	return vecL,vecH
end

function FA_SharedUtils:SelectedBonesCenter(moho,flag)
	local vecL,vecH= self:SelectedBonesBounds(moho,flag)
	return (vecL + vecH) / 2
end

function FA_SharedUtils:SelectedBonesBounds(moho,flag)
	local vecL = LM.Vector2:new_local()
	local vecH = LM.Vector2:new_local()
	local vecC = LM.Vector2:new_local()
	vecL:Set(0,0)
	vecH:Set(0,0)
	local cnt= nil

	local skel = moho:Skeleton()

	if (skel ~= nil) then
		local vec = LM.Vector2:new_local()
		local tMat = LM.Matrix:new_local()

		for i = 0, skel:CountBones() - 1 do
			local bone= skel:Bone(i)

			if (bone.fSelected) then
				vec:Set(0,0)

				if (moho.frame == 0) then
					tMat:Set(bone.fRestMatrix)
				else
					tMat:Set(bone.fMovedMatrix)
				end
				tMat:Transform(vec)

				if (cnt) then
					if (vecL.x > vec.x) then
						vecL.x= vec.x
					elseif (vecH.x < vec.x) then
						vecH.x= vec.x
					end
					if (vecL.y > vec.y) then
						vecL.y= vec.y
					elseif (vecH.y < vec.y) then
						vecH.y= vec.y
					end
					vecC= vecC + vec
					cnt= cnt + 1
				else
					vecL:Set(vec)
					vecH:Set(vec)
					vecC:Set(vec)
					cnt= 1
				end
				vec:Set(bone.fLength,0)
				tMat:Transform(vec)

				if (vecL.x > vec.x) then
					vecL.x= vec.x
				elseif (vecH.x < vec.x) then
					vecH.x= vec.x
				end
				if (vecL.y > vec.y) then
					vecL.y= vec.y
				elseif (vecH.y < vec.y) then
					vecH.y= vec.y
				end
				vecC= vecC + vec
			end
		end
	end

	if (cnt and flag) then
--center correction
		vecC= (vecC / cnt) - vecL - vecH

		if (vecC.x < 0) then
			vecL.x= vecL.x + vecC.x
		else
			vecH.x= vecH.x + vecC.x
		end

		if (vecC.y < 0) then
			vecL.y= vecL.y + vecC.y
		else
			vecH.y= vecH.y + vecC.y
		end
	end
	return vecL,vecH
end

function FA_SharedUtils:PrepMoveBones(moho)
	local skel= moho:Skeleton()

	if (skel == nil) then return end

	if (moho.frame == 0) then

		for i = 0, skel:CountBones() - 1 do
			local bone= skel:Bone(i)

			if (bone.fSelected) then
				bone.fTempPos:Set(bone.fAnimPos:GetValue(0))
				bone.fTempLength = bone.fLength
				bone.fTempAngle = bone.fAnimAngle:GetValue(0)
			elseif (bone.fParent >= 0) then

				if (skel:Bone(bone.fParent).fSelected) then
					bone.fTempPos:Set(bone.fAnimPos:GetValue(0))
					bone.fTempLength = bone.fLength
					bone.fTempAngle = bone.fAnimAngle:GetValue(0)
				end
			end
		end
	else
		for i = 0, skel:CountBones() - 1 do
			local bone= skel:Bone(i)

			if (bone.fSelected) then
				bone.fTempPos:Set(bone.fPos)
				bone.fTempLength = bone.fLength
				bone.fTempAngle = bone.fAngle
			elseif (bone.fParent >= 0) then

				if (skel:Bone(bone.fParent).fSelected) then
					bone.fTempPos:Set(bone.fPos)
					bone.fTempLength = bone.fLength
					bone.fTempAngle = bone.fAngle
				end
			end
		end
	end
end

function FA_SharedUtils:TranslateBones(moho,vec)
	local skel= moho:Skeleton()
	if (skel == nil) then return end

	local boneVec = LM.Vector2:new_local()
	local invM = LM.Matrix:new_local()

--pass1: inverse transform child bones

	for i= 0, skel:CountBones() - 1 do
		local bone= skel:Bone(i)

		if (not bone.fSelected and bone.fParent >= 0) then
			local pBone= skel:Bone(bone.fParent)

			if (pBone.fSelected) then

				if (moho.frame == 0) then
					invM:Set(pBone.fRestMatrix)
				else
					invM:Set(pBone.fMovedMatrix)
				end
				boneVec:Set(bone.fTempPos)
				invM:Transform(boneVec)
				boneVec= boneVec - vec
				invM:Invert()
				invM:Transform(boneVec)
				bone.fAnimPos:SetValue(moho.frame,boneVec)
			end
		end
	end

--pass2: transform bones

	for i= 0, skel:CountBones() - 1 do
		local bone= skel:Bone(i)

		if (bone.fSelected) then

			if (bone.fParent >= 0) then
				local pBone= skel:Bone(bone.fParent)

				if (not pBone.fSelected) then

					if (moho.frame == 0) then
						invM:Set(pBone.fRestMatrix)
					else
						invM:Set(pBone.fMovedMatrix)
					end
					boneVec:Set(bone.fTempPos)
					invM:Transform(boneVec)
					boneVec= boneVec + vec
					invM:Invert()
					invM:Transform(boneVec)
					bone.fAnimPos:SetValue(moho.frame,boneVec)
				end
			else
				boneVec= vec + bone.fTempPos
				bone.fAnimPos:SetValue(moho.frame,boneVec)
			end
		end
	end
end

function FA_SharedUtils:RotateBones(moho,angle,centerVec)
	local skel= moho:Skeleton()
	if (skel == nil) then return end

	local boneVec = LM.Vector2:new_local()
	local invM = LM.Matrix:new_local()
	local invPM = LM.Matrix:new_local()

--pass1: inverse transform child bones

	for i= 0, skel:CountBones() - 1 do
		local bone= skel:Bone(i)

		if (not bone.fSelected and bone.fParent >= 0) then
			local pBone= skel:Bone(bone.fParent)

			if (pBone.fSelected) then

				if (moho.frame == 0) then
					invM:Set(pBone.fRestMatrix)
				else
					invM:Set(pBone.fMovedMatrix)
				end
				boneVec:Set(bone.fTempPos)
				invM:Transform(boneVec)
				boneVec= boneVec - centerVec
				boneVec:Rotate(-angle)
				boneVec= boneVec + centerVec
				invM:Invert()
				invM:Transform(boneVec)
				bone.fAnimPos:SetValue(moho.frame,boneVec)
				bone.fAnimAngle:SetValue(moho.frame,bone.fTempAngle - angle)
			end
		end
	end

--pass2: transform bones

	for i= 0, skel:CountBones() - 1 do
		local bone= skel:Bone(i)

		if (bone.fSelected) then

			if (bone.fParent >= 0) then
				local pBone= skel:Bone(bone.fParent)

				if (not pBone.fSelected) then

					if (moho.frame == 0) then
						invM:Set(pBone.fRestMatrix)
					else
						invM:Set(pBone.fMovedMatrix)
					end
					boneVec:Set(bone.fTempPos)
					invM:Transform(boneVec)
					boneVec= boneVec - centerVec
					boneVec:Rotate(angle)
					boneVec= boneVec + centerVec
					invM:Invert()
					invM:Transform(boneVec)
					bone.fAnimPos:SetValue(moho.frame,boneVec)
					bone.fAnimAngle:SetValue(moho.frame,bone.fTempAngle + angle)
				end
			else
				boneVec= bone.fTempPos - centerVec
				boneVec:Rotate(angle)
				boneVec= boneVec + centerVec
				bone.fAnimPos:SetValue(moho.frame,boneVec)
				bone.fAnimAngle:SetValue(moho.frame,bone.fTempAngle + angle)
			end
		end
	end
end

function FA_SharedUtils:SelectItem(moho, mesh, m)
	moho.document:PrepUndo(moho.layer)
	mesh:SelectNone()
	mesh:SelectGroup(mesh:Group(m):Name())
	moho:UpdateUI()
end

function FA_SharedUtils:SelectBone(moho, skel, m, ptmode)
	mesh.document:PrepUndo(moho.layer)

	for i = 0, skel:CountBones() - 1 do
		skel:Bone(i).fSelected = (i == m)
	end

	if (ptmode) then
		moho:UpdateBonePointSelection()
	end
	moho:UpdateUI()
end
